testinfra_hosts = ['ansible://letsencryptstaging-host']


def test_can_curl_locally(host):
    with host.sudo():
        host.run("apt-get install -y curl")
    assert host.run("curl -m 5 -I https://letsencryptstaging-host.$(hostname -d)").rc == 0
    assert host.run("grep 'SOMETHING TESTS CAN GREP' /etc/nginx/sites-enabled/*.conf").rc == 0


def test_bind_fails_because_no_lestencrypt_staging_ca_installed(host):
    bind_host = host.get_host('ansible://bind-host',
                              ansible_inventory=host.backend.ansible_inventory)
    with bind_host.sudo():
        bind_host.run("apt-get install -y curl")
    assert bind_host.run("curl -m 5 -I -k https://letsencryptstaging-host.$(hostname -d)").rc == 0
    assert bind_host.run("curl -m 5 -I https://letsencryptstaging-host.$(hostname -d)").rc == 60


def test_client_succeeds_because_lestencrypt_staging_ca_installed(host):
    client_host = host.get_host('ansible://client-host',
                                ansible_inventory=host.backend.ansible_inventory)
    with client_host.sudo():
        client_host.run("apt-get install -y curl")
    assert client_host.run("curl -m 5 -I https://letsencryptstaging-host.$(hostname -d)").rc == 0


def test_reverse_can_curl_locally(host):
    with host.sudo():
        host.run("apt-get install -y curl")
    assert host.run("curl -m 5 -I https://reverse.$(hostname -d)").rc == 0


def test_reverse_file_behind_the_reverse_proxy_is_read(host):
    r = host.run("curl -s -m 5 https://reverse.$(hostname -d)/index.html")
    assert r.rc == 0
    assert 'REVERSEPROXY' in r.stdout


def test_reverse_bind_fails_because_no_lestencrypt_staging_ca_installed(host):
    bind_host = host.get_host('ansible://bind-host',
                              ansible_inventory=host.backend.ansible_inventory)
    with bind_host.sudo():
        bind_host.run("apt-get install -y curl")
    assert bind_host.run("curl -m 5 -I -k https://reverse.$(hostname -d)").rc == 0
    assert bind_host.run("curl -m 5 -I https://reverse.$(hostname -d)").rc == 60


def test_reverse_client_succeeds_because_lestencrypt_staging_ca_installed(host):
    client_host = host.get_host('ansible://client-host',
                                ansible_inventory=host.backend.ansible_inventory)
    with client_host.sudo():
        client_host.run("apt-get install -y curl")
    assert client_host.run("curl -m 5 -I https://reverse.$(hostname -d)").rc == 0



testinfra_hosts = ['jmm-host']


def test_nextcloud(host):
    cmd = host.run("""
    set -xe
    d=/dev/sdb
    test -e /dev/sdb || d=/dev/vda
    mount | grep $d | grep /var/lib/docker
    curl --silent https://jmm.$(hostname -d)/login | grep --quiet 'Forgot pass'
    """)
    print(cmd.stdout)
    print(cmd.stderr)
    assert 0 == cmd.rc


def test_nextcloud_via_tor(host):
    cmd = host.run("""
    set -xe
    hostname=$(sudo cat /var/lib/tor/services/cloud/hostname)
    torsocks curl --silent http://$hostname/login | grep --quiet 'Forgot pass'
    """)
    print(cmd.stdout)
    print(cmd.stderr)
    assert 0 == cmd.rc



from cliff.command import Command

from enough import settings
from enough.common import options
from enough.common import ssh


class SSH(Command):
    "SSH to host"

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument('host')
        parser.add_argument('args', nargs='*')
        return options.set_options(parser)

    def take_action(self, parsed_args):
        args = vars(self.app.options)
        args.update(vars(parsed_args))
        s = ssh.SSH(settings.CONFIG_DIR, **args)
        s.ssh(args['host'], args['args'])



# Copyright (C) 2015 Maciej Delmanowski <drybjed@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <https://www.gnu.org/licenses/>.

from ansible import errors

try:
    import fnmatch
except Exception as e:
    raise errors.AnsibleFilterError('fnmatch python library not found')


def globmatch_filter(value, pattern):
    ''' Return string or list of items matching given glob pattern(s). '''

    if not isinstance(pattern, (list, tuple)):
        pattern = [pattern]

    if isinstance(value, (list, tuple)):
        _ret = []

        for element in pattern:
            for entry in value:
                if fnmatch.fnmatch(str(entry), str(element)):
                    if entry not in _ret:
                        _ret.append(entry)

        if _ret:
            return _ret
        else:
            return list()

    else:

        for element in pattern:
            if fnmatch.fnmatch(str(value), str(element)):
                return value


class FilterModule(object):

    ''' Return string or list of items matching given glob pattern(s). '''
    def filters(self):
        return {
            'globmatch': globmatch_filter
        }



from cliff.command import Command

from enough import settings
from enough.common import Enough
from enough.common import options


class Resize(Command):
    "Resize a volume"

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument('host')
        parser.add_argument('volume')
        return options.set_options(parser)

    def take_action(self, parsed_args):
        args = vars(self.app.options)
        args.update(vars(parsed_args))
        e = Enough(settings.CONFIG_DIR, settings.SHARE_DIR, **args)
        e.volume_resize(args['host'], args['volume'])



# Copyright 2016 Pierre Chifflier <pollux@wzdftpd.net>
#
# SSH + lxc-attach connection module for Ansible 2.0
#
# Adapted from ansible/plugins/connection/ssh.py
# Forked from https://github.com/chifflier/ansible-lxc-ssh
# Hosted on https://github.com/andreasscherbaum/ansible-lxc-ssh
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
import errno
import fcntl
import hashlib
import os
import pipes
import pty
import shlex
import subprocess
import sys
import time
from distutils.version import LooseVersion

from ansible import constants as C
from ansible.utils.path import unfrackpath, makedirs_safe
from ansible.release import __version__ as ansible_version
if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
    from functools import wraps
if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
    from ansible.compat.six import text_type, binary_type
    from ansible.errors import AnsibleConnectionFailure, AnsibleError
    from ansible.plugins.connection import ConnectionBase
else:  # newer version
    from ansible.errors import (AnsibleError, AnsibleConnectionFailure,
                                AnsibleFileNotFound)
    from ansible.compat import selectors
    from ansible.module_utils.six import PY3, text_type, binary_type
    if LooseVersion(ansible_version) < LooseVersion('2.4.0.0'):
        from ansible.compat.six.moves import shlex_quote
    else:
        from ansible.module_utils.six.moves import shlex_quote
    from ansible.plugins.connection import ConnectionBase

if LooseVersion(ansible_version) >= LooseVersion('2.2.0.0'):
    from ansible.module_utils._text import to_bytes, to_text, to_native
else:
    from ansible.utils.unicode import (to_bytes, to_unicode as to_text,
                                       to_native)

try:
    from __main__ import display
except ImportError:
    from ansible.utils.display import Display
    display = Display()

if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
    import select


# only used from Ansible version 2.3 on forward
class AnsibleControlPersistBrokenPipeError(AnsibleError):
    ''' ControlPersist broken pipe '''
    pass


def _ssh_retry(func):
    """
    Decorator to retry ssh/scp/sftp in the case of a connection failure

    Will retry if:
    * an exception is caught
    * ssh returns 255
    Will not retry if
    * remaining_tries is <2
    * retries limit reached
    """
    @wraps(func)
    def wrapped(self, *args, **kwargs):
        remaining_tries = int(C.ANSIBLE_SSH_RETRIES) + 1
        cmd_summary = "%s..." % args[0]
        for attempt in range(remaining_tries):
            cmd = args[0]
            if (attempt != 0 and self._play_context.password
                    and isinstance(cmd, list)):
                # If this is a retry, the fd/pipe for sshpass is closed,
                # and we need a new one
                self.sshpass_pipe = os.pipe()
                cmd[1] = b'-d' + to_bytes(self.sshpass_pipe[0],
                                          nonstring='simplerepr',
                                          errors='surrogate_or_strict')

            try:
                try:
                    return_tuple = func(self, *args, **kwargs)
                    display.vvv(return_tuple, host=self.host)
                    # 0 = success
                    # 1-254 = remote command return code
                    # 255 = failure from the ssh command itself
                except (AnsibleControlPersistBrokenPipeError) as e:
                    # Retry one more time because of the ControlPersist
                    # broken pipe (see #16731)
                    display.vvv(u"RETRYING BECAUSE OF CONTROLPERSIST "
                                "BROKEN PIPE")
                    return_tuple = func(self, *args, **kwargs)

                if return_tuple[0] != 255:
                    break
                else:
                    raise AnsibleConnectionFailure(
                            "Failed to connect to the host via ssh: %s" %
                            to_native(return_tuple[2]))
            except (AnsibleConnectionFailure, Exception) as e:
                if attempt == remaining_tries - 1:
                    raise
                else:
                    pause = 2 ** attempt - 1
                    if pause > 30:
                        pause = 30

                    if isinstance(e, AnsibleConnectionFailure):
                        msg = (("ssh_retry: attempt: %d, ssh return code is "
                                "255. cmd (%s), pausing for %d seconds") %
                               (attempt, cmd_summary, pause))
                    else:
                        msg = (("ssh_retry: attempt: %d, caught exception(%s) "
                                "from cmd (%s), pausing for %d seconds") %
                               (attempt, e, cmd_summary, pause))

                    display.vv(msg, host=self.host)

                    time.sleep(pause)
                    continue

        return return_tuple
    return wrapped


class Connection(ConnectionBase):
    ''' ssh+lxc_attach connection '''
    transport = 'lxc_ssh'

    def __init__(self, play_context, new_stdin, *args, **kwargs):
        # print args
        # print kwargs
        super(Connection, self).__init__(play_context, new_stdin,
                                         *args, **kwargs)
        self.host = self._play_context.remote_addr
        if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
            self.port = self._play_context.port
            self.user = self._play_context.remote_user
            self.control_path = C.ANSIBLE_SSH_CONTROL_PATH
            self.control_path_dir = C.ANSIBLE_SSH_CONTROL_PATH_DIR
        self.lxc_version = None

        # LXC v1 uses 'lxc-info', 'lxc-attach' and so on
        # LXC v2 uses just 'lxc'
        (returncode2, stdout2, stderr2) = (
                self._exec_command("which lxc", None, False))
        (returncode1, stdout1, stderr1) = (
                self._exec_command("which lxc-info", None, False))
        if (returncode2 == 0):
            self.lxc_version = 2
            display.vvv('LXC v2')
        elif (returncode1 == 0):
            self.lxc_version = 1
            display.vvv('LXC v1')
        else:
            raise AnsibleConnectionFailure('Cannot identify LXC version')
            sys.exit(1)

    def _connect(self):
        # The connection is created by running ssh/scp/sftp from the
        # exec_command, put_file, and fetch_file methods, so we don't need
        # to do any connection management here.
        ''' connect to the lxc; nothing to do here '''
        display.vvv('XXX connect')
        super(Connection, self)._connect()
        # self.container_name = self.ssh._play_context.remote_addr
        self.container_name = self._play_context.ssh_extra_args  # XXX
        # self.container = None

    @staticmethod
    def _create_control_path(host, port, user, connection=None):
        # only used from Ansible version 2.3 on forward
        '''Make a hash for the controlpath based on con attributes'''
        pstring = '%s-%s-%s' % (host, port, user)
        if connection:
            pstring += '-%s' % connection
        m = hashlib.sha1()
        m.update(to_bytes(pstring))
        digest = m.hexdigest()
        cpath = '%(directory)s/' + digest[:10]
        return cpath

    @staticmethod
    def _persistence_controls(b_command):
        '''
        Takes a command array and scans it for ControlPersist and ControlPath
        settings and returns two booleans indicating whether either was found.
        This could be smarter, e.g. returning false if ControlPersist is 'no',
        but for now we do it simple way.
        '''

        controlpersist = False
        controlpath = False

        if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
            for arg in b_command:
                if 'controlpersist' in arg.lower():
                    controlpersist = True
                elif 'controlpath' in arg.lower():
                    controlpath = True
        if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
            for b_arg in (a.lower() for a in b_command):
                if b'controlpersist' in b_arg:
                    controlpersist = True
                elif b'controlpath' in b_arg:
                    controlpath = True

        return controlpersist, controlpath

    @staticmethod
    def _split_args(argstring):
        """
        Takes a string like '-o Foo=1 -o Bar="foo bar"' and returns a
        list ['-o', 'Foo=1', '-o', 'Bar=foo bar'] that can be added to
        the argument list. The list will not contain any empty elements.
        """
        return ([to_text(x.strip())
                for x in shlex.split(argstring) if x.strip()])

    if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
        def _add_args(self, explanation, args):
            """
            Adds the given args to self._command and displays a caller-supplied
            explanation of why they were added.
            """
            self._command += args
            display.vvvvv('SSH: ' + explanation + ': (%s)' %
                          ')('.join(args), host=self._play_context.remote_addr)

    if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
        def _add_args(self, b_command, b_args, explanation):
            """
            Adds arguments to the ssh command and displays a caller-supplied
            explanation of why.
            :arg b_command: A list containing the command to add the new
                arguments to. This list will be modified by this method.
            :arg b_args: An iterable of new arguments to add. This iterable
                is used more than once so it must be persistent
                (ie: a list is okay but a StringIO would not)
            :arg explanation: A text string containing explaining why
                the arguments were added. It will be displayed with
                a high enough verbosity.
            .. note:: This function does its work via side-effect.
               The b_command list has the new arguments appended.
            """
            display.vvvvv(u'SSH: %s: (%s)' % (
                explanation, ')('.join(to_text(a) for a in b_args)),
                host=self._play_context.remote_addr)
            b_command += b_args

    if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
        def _build_command(self, binary, *other_args):
            self._command = []
            self._command += [binary]
            self._command += ['-C']
            if self._play_context.verbosity > 3:
                self._command += ['-vvv']
            elif binary == 'ssh':
                # Older versions of ssh (e.g. in RHEL 6) don't accept sftp -q.
                self._command += ['-q']
            # Next, we add [ssh_connection]ssh_args from ansible.cfg.
            if self._play_context.ssh_args:
                args = self._split_args(self._play_context.ssh_args)
                self._add_args("ansible.cfg set ssh_args", args)
            # Now we add various arguments controlled by configuration file
            # settings (e.g. host_key_checking) or inventory variables
            # (ansible_ssh_port) or a combination thereof.
            if not C.HOST_KEY_CHECKING:
                self._add_args(
                    "ANSIBLE_HOST_KEY_CHECKING/host_key_checking disabled",
                    ("-o", "StrictHostKeyChecking=no")
                )
            if self._play_context.port is not None:
                self._add_args(
                    "ANSIBLE_REMOTE_PORT/remote_port/ansible_port set",
                    ("-o", "Port={0}".format(self._play_context.port))
                )
            key = self._play_context.private_key_file
            if key:
                self._add_args(
                    ("ANSIBLE_PRIVATE_KEY_FILE/private_key_file/"
                     "ansible_ssh_private_key_file set"),
                    ("-o", "IdentityFile=\"{0}\"".format(
                        os.path.expanduser(key)))
                )
            if not self._play_context.password:
                self._add_args(
                    "ansible_password/ansible_ssh_pass not set", (
                        "-o", "KbdInteractiveAuthentication=no",
                        "-o", ("PreferredAuthentications=gssapi-with-mic,"
                               "gssapi-keyex,hostbased,publickey"),
                        "-o", "PasswordAuthentication=no"
                    )
                )
            user = self._play_context.remote_user
            if user:
                self._add_args(
                    "ANSIBLE_REMOTE_USER/remote_user/ansible_user/user/-u set",
                    ("-o", "User={0}".format(
                        to_bytes(self._play_context.remote_user)))
                )
            self._add_args(
                "ANSIBLE_TIMEOUT/timeout set",
                ("-o", "ConnectTimeout={0}".format(self._play_context.timeout))
            )
            # Check if ControlPersist is enabled and add a ControlPath
            # if one hasn't already been set.
            controlpersist, controlpath = (
                    self._persistence_controls(self._command))
            if controlpersist:
                self._persistent = True
                if not controlpath:
                    cpdir = unfrackpath('$HOME/.ansible/cp')
                    display.vv(str(C.ANSIBLE_SSH_CONTROL_PATH))
                    # The directory must exist and be writable.
                    makedirs_safe(cpdir, 0o700)
                    if not os.access(cpdir, os.W_OK):
                        raise AnsibleError("Cannot write to ControlPath %s" %
                                           cpdir)
                    args = ("-o", "ControlPath={0}".format(
                            to_bytes(C.ANSIBLE_SSH_CONTROL_PATH %
                                     dict(directory=cpdir)))
                            )
                    self._add_args(("found only ControlPersist; "
                                    "added ControlPath"), args)
            # Finally, we add any caller-supplied extras.
            if other_args:
                self._command += other_args
            return self._command

    if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
        def _build_command(self, binary, *other_args):
            b_command = []
            if binary == 'ssh':
                b_command += [to_bytes(self._play_context.ssh_executable,
                                       errors='surrogate_or_strict')]
            else:
                b_command += [to_bytes(binary, errors='surrogate_or_strict')]
            if self._play_context.verbosity > 3:
                b_command.append(b'-vvv')
            # Next, we add [ssh_connection]ssh_args from ansible.cfg.
            #
            if self._play_context.ssh_args:
                b_args = [to_bytes(a, errors='surrogate_or_strict') for a in
                          self._split_args(self._play_context.ssh_args)]
                self._add_args(b_command, b_args, u"ansible.cfg set ssh_args")

            # Now we add various arguments controlled by configuration file
            # settings (e.g. host_key_checking) or inventory variables
            # (ansible_ssh_port) or a combination thereof.
            if not C.HOST_KEY_CHECKING:
                b_args = (b"-o", b"StrictHostKeyChecking=no")
                self._add_args(b_command, b_args,
                               (u"ANSIBLE_HOST_KEY_CHECKING/host_key_checking "
                                "disabled"))
            if self._play_context.port is not None:
                b_args = (b"-o", b"Port=" +
                          to_bytes(self._play_context.port,
                                   nonstring='simplerepr',
                                   errors='surrogate_or_strict'))
                self._add_args(b_command, b_args,
                               (u"ANSIBLE_REMOTE_PORT/remote_port/"
                                "ansible_port set"))
            key = self._play_context.private_key_file
            if key:
                b_args = (b"-o", b'IdentityFile="' +
                          to_bytes(os.path.expanduser(key),
                                   errors='surrogate_or_strict') + b'"')
                self._add_args(b_command, b_args,
                               (u"ANSIBLE_PRIVATE_KEY_FILE/private_key_file/"
                                "ansible_ssh_private_key_file set"))
            if not self._play_context.password:
                self._add_args(
                    b_command, (
                        b"-o", b"KbdInteractiveAuthentication=no",
                        b"-o", (b"PreferredAuthentications=gssapi-with-mic,"
                                b"gssapi-keyex,hostbased,publickey"),
                        b"-o", b"PasswordAuthentication=no"
                    ),
                    u"ansible_password/ansible_ssh_pass not set"
                )
            user = self._play_context.remote_user
            if user:
                self._add_args(
                    b_command,
                    (b"-o", b"User=" +
                     to_bytes(self._play_context.remote_user,
                              errors='surrogate_or_strict')),
                    u"ANSIBLE_REMOTE_USER/remote_user/ansible_user/user/-u set"
                )
            self._add_args(
                b_command,
                (b"-o", b"ConnectTimeout=" +
                 to_bytes(self._play_context.timeout,
                          errors='surrogate_or_strict',
                          nonstring='simplerepr')),
                u"ANSIBLE_TIMEOUT/timeout set"
            )
            # Check if ControlPersist is enabled and add a ControlPath
            # if one hasn't already been set.
            controlpersist, controlpath = self._persistence_controls(b_command)
            if controlpersist:
                self._persistent = True
                if not controlpath:
                    cpdir = unfrackpath(self.control_path_dir)
                    b_cpdir = to_bytes(cpdir, errors='surrogate_or_strict')
                    # The directory must exist and be writable.
                    makedirs_safe(b_cpdir, 0o700)
                    if not os.access(b_cpdir, os.W_OK):
                        raise AnsibleError(
                                ("Cannot write to ControlPath %s" %
                                 to_native(cpdir)))

                    if not self.control_path:
                        self.control_path = self._create_control_path(
                            self.host,
                            self.port,
                            self.user
                        )
                    b_args = (b"-o", b"ControlPath=" +
                              to_bytes(self.control_path %
                                       dict(directory=cpdir),
                                       errors='surrogate_or_strict'))
                    self._add_args(b_command, b_args,
                                   (u"found only ControlPersist; "
                                    "added ControlPath"))

            # Finally, we add any caller-supplied extras.
            if other_args:
                b_command += [to_bytes(a) for a in other_args]

            return b_command

    def _send_initial_data(self, fh, in_data):
        '''
        Writes initial data to the stdin filehandle of the subprocess
        and closes it. (The handle must be closed; otherwise, for example,
        "sftp -b -" will just hang forever waiting for more commands.)
        '''

        display.debug('Sending initial data')

        try:
            if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
                fh.write(in_data)
            if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
                fh.write(to_bytes(in_data))
            fh.close()
        except (OSError, IOError):
            raise AnsibleConnectionFailure(
                    ('SSH Error: data could not be sent to remote host '
                     '"%s". Make sure this host can be reached over '
                     'ssh') % self.host)

        display.debug('Sent initial data (%d bytes)' % len(in_data))

    @staticmethod
    def _terminate_process(p):
        # Used by _run() to kill processes on failures
        """ Terminate a process, ignoring errors """
        try:
            p.terminate()
        except (OSError, IOError):
            pass

    if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
        # This is separate from _run() because we need to do the same thing
        # for stdout and stderr.
        def _examine_output(self, source, state, chunk, sudoable):
            '''
            Takes a string, extracts complete lines from it, tests to see
            if they are a prompt, error message, etc., and sets appropriate
            flags in self. Prompt and success lines are removed.

            Returns the processed (i.e. possibly-edited) output and the
            unprocessed remainder (to be processed with the next chunk)
            as strings.
            '''

            output = []
            for line in chunk.splitlines(True):
                suppress_output = False

                # display.debug(
                #         ("Examining line (source=%s, state=%s): "
                #          "'%s'") % (source, state, l.rstrip('\r\n')))
                if (self._play_context.prompt
                        and self.check_password_prompt(line)):
                    display.debug(
                            ("become_prompt: (source=%s, state=%s): "
                             "'%s'") % (source, state, line.rstrip('\r\n')))
                    self._flags['become_prompt'] = True
                    suppress_output = True
                elif (self._play_context.success_key
                      and self.check_become_success(line)):
                    display.debug(
                            ("become_success: (source=%s, state=%s): "
                             "'%s'") % (source, state, line.rstrip('\r\n')))
                    self._flags['become_success'] = True
                    suppress_output = True
                elif sudoable and self.check_incorrect_password(line):
                    display.debug(
                            ("become_error: (source=%s, state=%s): "
                             "'%s'") % (source, state, line.rstrip('\r\n')))
                    self._flags['become_error'] = True
                elif sudoable and self.check_missing_password(line):
                    display.debug(
                            ("become_nopasswd_error: (source=%s, state=%s): "
                             "'%s'") % (source, state, line.rstrip('\r\n')))
                    self._flags['become_nopasswd_error'] = True

                if not suppress_output:
                    output.append(line)

            # The chunk we read was most likely a series of complete lines,
            # but just in case the last line was incomplete (and not a prompt,
            # which we would have removed from the output), we retain it to be
            # processed with the next chunk.

            remainder = ''
            if output and not output[-1].endswith('\n'):
                remainder = output[-1]
                output = output[:-1]

            return ''.join(output), remainder

    if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
        # This is separate from _run() because we need to do the same thing
        # for stdout and stderr.
        def _examine_output(self, source, state, b_chunk, sudoable):
            '''
            Takes a string, extracts complete lines from it, tests to see
            if they are a prompt, error message, etc., and sets appropriate
            flags in self. Prompt and success lines are removed.

            Returns the processed (i.e. possibly-edited) output and the
            unprocessed remainder (to be processed with the next chunk)
            as strings.
            '''

            output = []
            for b_line in b_chunk.splitlines(True):
                display_line = to_text(b_line).rstrip('\r\n')
                suppress_output = False

                # display.debug(
                #         ("Examining line (source=%s, state=%s): "
                #          "'%s'") % (source, state, display_line))
                if (self._play_context.prompt
                        and self.check_password_prompt(b_line)):
                    display.debug(
                            ("become_prompt: (source=%s, state=%s): "
                             "'%s'") % (source, state, display_line))
                    self._flags['become_prompt'] = True
                    suppress_output = True
                elif (self._play_context.success_key
                      and self.check_become_success(b_line)):
                    display.debug(
                            ("become_success: (source=%s, state=%s): "
                             "'%s'") % (source, state, display_line))
                    self._flags['become_success'] = True
                    suppress_output = True
                elif sudoable and self.check_incorrect_password(b_line):
                    display.debug(
                            ("become_error: (source=%s, state=%s): "
                             "'%s'") % (source, state, display_line))
                    self._flags['become_error'] = True
                elif sudoable and self.check_missing_password(b_line):
                    display.debug(
                            ("become_nopasswd_error: (source=%s, state=%s): "
                             "'%s'") % (source, state, display_line))
                    self._flags['become_nopasswd_error'] = True

                if not suppress_output:
                    output.append(b_line)

            # The chunk we read was most likely a series of complete lines,
            # but just in case the last line was incomplete (and not a prompt,
            # which we would have removed from the output), we retain it to be
            # processed with the next chunk.

            remainder = b''
            if output and not output[-1].endswith(b'\n'):
                remainder = output[-1]
                output = output[:-1]

            return b''.join(output), remainder

    def _bare_run(self, cmd, in_data, sudoable=True, checkrc=True):
        # only used from Ansible version 2.3 on forward
        '''
        Starts the command and communicates with it until it ends.
        '''

        display_cmd = list(map(shlex_quote, map(to_text, cmd)))
        display.vvv(u'SSH: EXEC {0}'.format(u' '.join(display_cmd)),
                    host=self.host)

        # Start the given command. If we don't need to pipeline data,
        # we can try to use a pseudo-tty (ssh will have been invoked with -tt).
        # If we are pipelining data, or can't create a pty, we fall back to
        # using plain old pipes.

        p = None

        if isinstance(cmd, (text_type, binary_type)):
            cmd = to_bytes(cmd)
        else:
            cmd = list(map(to_bytes, cmd))

        if not in_data:
            try:
                # Make sure stdin is a proper pty to avoid tcgetattr errors
                master, slave = pty.openpty()
                if PY3 and self._play_context.password:
                    p = subprocess.Popen(cmd, stdin=slave,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE,
                                         pass_fds=self.sshpass_pipe)
                else:
                    p = subprocess.Popen(cmd, stdin=slave,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)
                stdin = os.fdopen(master, 'wb', 0)
                os.close(slave)
            except (OSError, IOError):
                p = None

        if not p:
            if PY3 and self._play_context.password:
                p = subprocess.Popen(cmd, stdin=subprocess.PIPE,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE,
                                     pass_fds=self.sshpass_pipe)
            else:
                p = subprocess.Popen(cmd, stdin=subprocess.PIPE,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)
            stdin = p.stdin

        # If we are using SSH password authentication, write the password into
        # the pipe we opened in _build_command.

        if self._play_context.password:
            os.close(self.sshpass_pipe[0])
            try:
                os.write(self.sshpass_pipe[1],
                         to_bytes(self._play_context.password) + b'\n')
            except OSError as e:
                # Ignore broken pipe errors if the sshpass process has exited.
                if e.errno != errno.EPIPE or p.poll() is None:
                    raise
            os.close(self.sshpass_pipe[1])

        #
        # SSH state machine
        #

        # Now we read and accumulate output from the running process until it
        # exits. Depending on the circumstances, we may also need to write an
        # escalation password and/or pipelined input to the process.

        states = [
            'awaiting_prompt', 'awaiting_escalation',
            'ready_to_send', 'awaiting_exit'
        ]

        # Are we requesting privilege escalation? Right now, we may be invoked
        # to execute sftp/scp with sudoable=True, but we can request escalation
        # only when using ssh. Otherwise we can send initial data straightaway.

        state = states.index('ready_to_send')
        if b'ssh' in cmd:
            if self._play_context.prompt:
                # We're requesting escalation with a password, so we have to
                # wait for a password prompt.
                state = states.index('awaiting_prompt')
                display.debug(u'Initial state: %s: %s' %
                              (states[state], self._play_context.prompt))
            elif self._play_context.become and self._play_context.success_key:
                # We're requesting escalation without a password, so we have to
                # detect success/failure before sending any initial data.
                state = states.index('awaiting_escalation')
                display.debug(u'Initial state: %s: %s' %
                              (states[state], self._play_context.success_key))

        # We store accumulated stdout and stderr output from the process here,
        # but strip any privilege escalation prompt/confirmation lines first.
        # Output is accumulated into tmp_*, complete lines are extracted into
        # an array, then checked and removed or copied to stdout or stderr. We
        # set any flags based on examining the output in self._flags.

        b_stdout = b_stderr = b''
        b_tmp_stdout = b_tmp_stderr = b''

        self._flags = dict(
            become_prompt=False, become_success=False,
            become_error=False, become_nopasswd_error=False
        )

        # select timeout should be longer than the connect timeout, otherwise
        # they will race each other when we can't connect, and the connect
        # timeout usually fails
        timeout = 2 + self._play_context.timeout
        for fd in (p.stdout, p.stderr):
            fcntl.fcntl(fd, fcntl.F_SETFL, fcntl.fcntl(fd, fcntl.F_GETFL)
                        | os.O_NONBLOCK)

        # TODO: bcoca would like to use SelectSelector() when open
        # filehandles is low, then switch to more efficient ones when higher.
        # select is faster when filehandles is low.
        selector = selectors.DefaultSelector()
        selector.register(p.stdout, selectors.EVENT_READ)
        selector.register(p.stderr, selectors.EVENT_READ)

        # If we can send initial data without waiting for anything, we do so
        # before we start polling
        if states[state] == 'ready_to_send' and in_data:
            self._send_initial_data(stdin, in_data)
            state += 1

        try:
            while True:
                poll = p.poll()
                events = selector.select(timeout)

                # We pay attention to timeouts only while negotiating a prompt.

                if not events:
                    # We timed out
                    if state <= states.index('awaiting_escalation'):
                        # If the process has already exited, then it's not
                        # really a timeout; we'll let the normal error
                        # handling deal with it.
                        if poll is not None:
                            break
                        self._terminate_process(p)
                        raise AnsibleError(
                                ('Timeout (%ds) waiting for privilege '
                                 'escalation prompt: %s') %
                                (timeout, to_native(b_stdout)))

                # Read whatever output is available on stdout and stderr,
                # and stop listening to the pipe if it's been closed.

                for key, event in events:
                    if key.fileobj == p.stdout:
                        b_chunk = p.stdout.read()
                        if b_chunk == b'':
                            # stdout has been closed, stop watching it
                            selector.unregister(p.stdout)
                            # When ssh has ControlMaster
                            # (+ControlPath/Persist) enabled, the first
                            # connection goes into the background and we
                            # never see EOF on stderr. If we see EOF on
                            # stdout, lower the select timeout to reduce
                            # the time wasted selecting on stderr if we
                            # observe that the process has not yet existed
                            # after this EOF. Otherwise we may spend a long
                            # timeout period waiting for an EOF that is
                            # not going to arrive until the persisted
                            # connection closes.
                            timeout = 1
                        b_tmp_stdout += b_chunk
                        display.debug(
                                ("stdout chunk (state=%s):\n>>>%s<<<\n" %
                                 (state, to_text(b_chunk))))
                    elif key.fileobj == p.stderr:
                        b_chunk = p.stderr.read()
                        if b_chunk == b'':
                            # stderr has been closed, stop watching it
                            selector.unregister(p.stderr)
                        b_tmp_stderr += b_chunk
                        display.debug(
                                ("stderr chunk (state=%s):\n>>>%s<<<\n" %
                                 (state, to_text(b_chunk))))

                # We examine the output line-by-line until we have negotiated
                # any privilege escalation prompt and subsequent success/error
                # message. Afterwards, we can accumulate output without
                # looking at it.

                if state < states.index('ready_to_send'):
                    if b_tmp_stdout:
                        b_output, b_unprocessed = (
                                self._examine_output('stdout', states[state],
                                                     b_tmp_stdout, sudoable))
                        b_stdout += b_output
                        b_tmp_stdout = b_unprocessed

                    if b_tmp_stderr:
                        b_output, b_unprocessed = (
                                self._examine_output('stderr', states[state],
                                                     b_tmp_stderr, sudoable))
                        b_stderr += b_output
                        b_tmp_stderr = b_unprocessed
                else:
                    b_stdout += b_tmp_stdout
                    b_stderr += b_tmp_stderr
                    b_tmp_stdout = b_tmp_stderr = b''

                # If we see a privilege escalation prompt,
                # we send the password. (If we're expecting a prompt but
                # the escalation succeeds, we didn't need the password and
                # can carry on regardless.)

                if states[state] == 'awaiting_prompt':
                    if self._flags['become_prompt']:
                        display.debug(
                                'Sending become_pass in response to prompt')
                        stdin.write(to_bytes(self._play_context.become_pass)
                                    + b'\n')
                        self._flags['become_prompt'] = False
                        state += 1
                    elif self._flags['become_success']:
                        state += 1

                # We've requested escalation (with or without a password),
                # now we wait for an error message or a successful escalation.

                if states[state] == 'awaiting_escalation':
                    if self._flags['become_success']:
                        display.debug('Escalation succeeded')
                        self._flags['become_success'] = False
                        state += 1
                    elif self._flags['become_error']:
                        display.debug('Escalation failed')
                        self._terminate_process(p)
                        self._flags['become_error'] = False
                        raise AnsibleError(
                                ('Incorrect %s password' %
                                 self._play_context.become_method))
                    elif self._flags['become_nopasswd_error']:
                        display.debug('Escalation requires password')
                        self._terminate_process(p)
                        self._flags['become_nopasswd_error'] = False
                        raise AnsibleError(
                                ('Missing %s password' %
                                 self._play_context.become_method))
                    elif self._flags['become_prompt']:
                        # This shouldn't happen, because we should see the
                        # "Sorry, try again" message first.
                        display.debug('Escalation prompt repeated')
                        self._terminate_process(p)
                        self._flags['become_prompt'] = False
                        raise AnsibleError(
                                ('Incorrect %s password' %
                                 self._play_context.become_method))

                # Once we're sure that the privilege escalation prompt,
                # if any, has been dealt with, we can send any initial data
                # and start waiting for output.

                if states[state] == 'ready_to_send':
                    if in_data:
                        self._send_initial_data(stdin, in_data)
                    state += 1

                # Now we're awaiting_exit: has the child process exited?
                # If it has, and we've read all available output from it,
                # we're done.

                if poll is not None:
                    if not selector.get_map() or not events:
                        break
                    # We should not see further writes to the stdout/stderr
                    # file descriptors after the process has closed, set the
                    # select timeout to gather any last writes we may have
                    # missed.
                    timeout = 0
                    continue

                # If the process has not yet exited, but we've already read
                # EOF from its stdout and stderr (and thus no longer
                # watching any file descriptors), we can just wait for it
                # to exit.

                elif not selector.get_map():
                    p.wait()
                    break

                # Otherwise there may still be outstanding data to read.
        finally:
            selector.close()
            # close stdin after process is terminated and stdout/stderr
            # are read completely (see also issue #848)
            stdin.close()

        if C.HOST_KEY_CHECKING:
            if cmd[0] == b"sshpass" and p.returncode == 6:
                raise AnsibleError(
                        ('Using a SSH password instead of a key is not '
                         'possible because Host Key checking is enabled and '
                         'sshpass does not support this.  Please add this '
                         'host\'s fingerprint to your known_hosts file to '
                         'manage this host.'))

        controlpersisterror = (b'Bad configuration option: ControlPersist'
                               in b_stderr
                               or (b'unknown configuration option: '
                                   b'ControlPersist') in b_stderr)
        if p.returncode != 0 and controlpersisterror:
            raise AnsibleError(
                    ('using -c ssh on certain older ssh versions may not '
                     'support ControlPersist, set ANSIBLE_SSH_ARGS="" '
                     '(or ssh_args in [ssh_connection] section of the '
                     'config file) before running again'))

        # If we find a broken pipe because of ControlPersist timeout
        # expiring (see #16731), we raise a special exception so that we
        # can retry a connection.
        controlpersist_broken_pipe = (
                (b'mux_client_hello_exchange: write packet: '
                 b'Broken pipe') in b_stderr)
        if p.returncode == 255 and controlpersist_broken_pipe:
            raise AnsibleControlPersistBrokenPipeError(
                    ('SSH Error: data could not be sent because of '
                     'ControlPersist broken pipe.'))

        if p.returncode == 255 and in_data and checkrc:
            raise AnsibleConnectionFailure(
                    ('SSH Error: data could not be sent to remote host "%s". '
                     'Make sure this host can be reached over ssh') %
                    self.host)

        return (p.returncode, b_stdout, b_stderr)

    if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
        @_ssh_retry
        def _run(self, cmd, in_data, sudoable=True, checkrc=True):
            """Wrapper around _bare_run that retries the connection
            """
            return self._bare_run(cmd, in_data, sudoable, checkrc)
    else:
        def _run(self, cmd, in_data, sudoable=True):
            '''
            Starts the command and communicates with it until it ends.
            '''

            display_cmd = map(to_text, map(pipes.quote, cmd))
            display.vvv(u'SSH: EXEC {0}'.format(u' '.join(display_cmd)),
                        host=self.host)

            # Start the given command. If we don't need to pipeline data,
            # we can try to use a pseudo-tty (ssh will have been invoked
            # with -tt). If we are pipelining data, or can't create a pty,
            # we fall back to using plain old pipes.

            p = None

            if isinstance(cmd, (text_type, binary_type)):
                cmd = to_bytes(cmd)
            else:
                cmd = list(map(to_bytes, cmd))

            if not in_data:
                try:
                    # Make sure stdin is a proper pty to avoid tcgetattr errors
                    master, slave = pty.openpty()
                    p = subprocess.Popen(cmd, stdin=slave,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)
                    stdin = os.fdopen(master, 'w', 0)
                    os.close(slave)
                except (OSError, IOError):
                    p = None

            if not p:
                p = subprocess.Popen(cmd, stdin=subprocess.PIPE,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)
                stdin = p.stdin

            # If we are using SSH password authentication,
            # write the password into the pipe we opened in _build_command.

            if self._play_context.password:
                os.close(self.sshpass_pipe[0])
                os.write(self.sshpass_pipe[1],
                         ("{0}\n".format(to_bytes(
                                         self._play_context.password))))
                os.close(self.sshpass_pipe[1])

            # SSH state machine
            #
            # Now we read and accumulate output from the running process
            # until it exits. Depending on the circumstances, we may also
            # need to write an escalation password and/or pipelined input
            # to the process.

            states = [
                'awaiting_prompt', 'awaiting_escalation',
                'ready_to_send', 'awaiting_exit'
            ]

            # Are we requesting privilege escalation? Right now, we may be
            # invoked to execute sftp/scp with sudoable=True, but we can
            # request escalation only when using ssh. Otherwise we can send
            # initial data straightaway.

            state = states.index('ready_to_send')
            if b'ssh' in cmd:
                if self._play_context.prompt:
                    # We're requesting escalation with a password,
                    # so we have to wait for a password prompt.
                    state = states.index('awaiting_prompt')
                    display.debug(
                            ('Initial state: %s: %s' %
                             (states[state], self._play_context.prompt)))
                elif (self._play_context.become
                      and self._play_context.success_key):
                    # We're requesting escalation without a password,
                    # so we have to detect success/failure before sending
                    # any initial data.
                    state = states.index('awaiting_escalation')
                    display.debug(
                            ('Initial state: %s: %s' %
                             (states[state], self._play_context.success_key)))

            # We store accumulated stdout and stderr output from the
            # process here, but strip any privilege escalation
            # prompt/confirmation lines first.
            # Output is accumulated into tmp_*, complete lines are extracted
            # into an array, then checked and removed or copied to stdout
            # or stderr. We set any flags based on examining the output
            # in self._flags.

            stdout = stderr = ''
            tmp_stdout = tmp_stderr = ''

            self._flags = dict(
                become_prompt=False, become_success=False,
                become_error=False, become_nopasswd_error=False
            )

            # select timeout should be longer than the connect timeout,
            # otherwise they will race each other when we can't connect,
            # and the connect timeout usually fails
            timeout = 2 + self._play_context.timeout
            rpipes = [p.stdout, p.stderr]
            for fd in rpipes:
                fcntl.fcntl(fd, fcntl.F_SETFL, fcntl.fcntl(fd, fcntl.F_GETFL)
                            | os.O_NONBLOCK)

            # If we can send initial data without waiting for anything,
            # we do so before we call select.

            if states[state] == 'ready_to_send' and in_data:
                self._send_initial_data(stdin, in_data)
                state += 1

            while True:
                rfd, wfd, efd = select.select(rpipes, [], [], timeout)

                # We pay attention to timeouts only while negotiating a prompt.

                if not rfd:
                    if state <= states.index('awaiting_escalation'):
                        # If the process has already exited, then it's not
                        # really a timeout; we'll let the normal error
                        # handling deal with it.
                        if p.poll() is not None:
                            break
                        self._terminate_process(p)
                        raise AnsibleError(
                                ('Timeout (%ds) waiting for privilege '
                                 'escalation prompt: %s' % (timeout, stdout)))

                # Read whatever output is available on stdout and stderr,
                # and stop listening to the pipe if it's been closed.

                if p.stdout in rfd:
                    chunk = p.stdout.read()
                    if chunk == '':
                        rpipes.remove(p.stdout)
                    tmp_stdout += chunk
                    display.debug(
                            ("stdout chunk (state=%s):\n>>>%s<<<\n" %
                             (state, chunk)))

                if p.stderr in rfd:
                    chunk = p.stderr.read()
                    if chunk == '':
                        rpipes.remove(p.stderr)
                    tmp_stderr += chunk
                    display.debug(
                            ("stderr chunk (state=%s):\n>>>%s<<<\n" %
                             (state, chunk)))

                # We examine the output line-by-line until we have negotiated
                # any privilege escalation prompt and subsequent success/error
                # message. Afterwards, we can accumulate output without
                # looking at it.

                if state < states.index('ready_to_send'):
                    if tmp_stdout:
                        output, unprocessed = (
                                self._examine_output('stdout', states[state],
                                                     tmp_stdout, sudoable))
                        stdout += output
                        tmp_stdout = unprocessed

                    if tmp_stderr:
                        output, unprocessed = (
                                self._examine_output('stderr', states[state],
                                                     tmp_stderr, sudoable))
                        stderr += output
                        tmp_stderr = unprocessed
                else:
                    stdout += tmp_stdout
                    stderr += tmp_stderr
                    tmp_stdout = tmp_stderr = ''

                # If we see a privilege escalation prompt, we send
                # the password. (If we're expecting a prompt but the
                # escalation succeeds, we didn't need the password and
                # can carry on regardless.)

                if states[state] == 'awaiting_prompt':
                    if self._flags['become_prompt']:
                        display.debug('Sending become_pass '
                                      'in response to prompt')
                        stdin.write('{0}\n'.format(
                                to_bytes(self._play_context.become_pass)))
                        self._flags['become_prompt'] = False
                        state += 1
                    elif self._flags['become_success']:
                        state += 1

                # We've requested escalation (with or without a password),
                # now we wait for an error message or a successful escalation.

                if states[state] == 'awaiting_escalation':
                    if self._flags['become_success']:
                        display.debug('Escalation succeeded')
                        self._flags['become_success'] = False
                        state += 1
                    elif self._flags['become_error']:
                        display.debug('Escalation failed')
                        self._terminate_process(p)
                        self._flags['become_error'] = False
                        raise AnsibleError(('Incorrect %s password' %
                                            self._play_context.become_method))
                    elif self._flags['become_nopasswd_error']:
                        display.debug('Escalation requires password')
                        self._terminate_process(p)
                        self._flags['become_nopasswd_error'] = False
                        raise AnsibleError(('Missing %s password' %
                                            self._play_context.become_method))
                    elif self._flags['become_prompt']:
                        # This shouldn't happen, because we should see the
                        # "Sorry, try again" message first.
                        display.debug('Escalation prompt repeated')
                        self._terminate_process(p)
                        self._flags['become_prompt'] = False
                        raise AnsibleError(
                                ('Incorrect %s password' %
                                    self._play_context.become_method))

                # Once we're sure that the privilege escalation prompt,
                # if any, has been dealt with, we can send any initial data
                # and start waiting for output.

                if states[state] == 'ready_to_send':
                    if in_data:
                        self._send_initial_data(stdin, in_data)
                    state += 1

                # Now we're awaiting_exit: has the child process exited?
                # If it has, and we've read all available output from it,
                # we're done.

                if p.poll() is not None:
                    if not rpipes or not rfd:
                        break

                    # When ssh has ControlMaster (+ControlPath/Persist)
                    # enabled, the first connection goes into the background
                    # and we never see EOF on stderr. If we see EOF on stdout
                    # and the process has exited, we're probably done.
                    # We call select again with a zero timeout, just to make
                    # certain we don't miss anything that may have been
                    # written to stderr between the time we called select()
                    # and when we learned that the process had finished.

                    if p.stdout not in rpipes:
                        timeout = 0
                        continue

                # If the process has not yet exited, but we've already read
                # EOF from its stdout and stderr (and thus removed both from
                # rpipes), we can just wait for it to exit.

                elif not rpipes:
                    p.wait()
                    break

                # Otherwise there may still be outstanding data to read.

            # close stdin after process is terminated and stdout/stderr are
            # read completely (see also issue #848)
            stdin.close()

            if C.HOST_KEY_CHECKING:
                if cmd[0] == b"sshpass" and p.returncode == 6:
                    raise AnsibleError(
                            ('Using a SSH password instead of a key '
                             'is not possible because Host Key checking '
                             'is enabled and sshpass does not support this. '
                             'Please add this host\'s fingerprint to your '
                             'known_hosts file to manage this host.'))

            controlpersisterror = ('Bad configuration option: ControlPersist'
                                   in stderr or
                                   ('unknown configuration option: '
                                    'ControlPersist') in stderr)
            if p.returncode != 0 and controlpersisterror:
                raise AnsibleError(
                        ('using -c ssh on certain older ssh versions '
                         'may not support ControlPersist, '
                         'set ANSIBLE_SSH_ARGS="" (or ssh_args in '
                         '[ssh_connection] section of the config file) '
                         'before running again'))

            if p.returncode == 255 and in_data:
                raise AnsibleConnectionFailure(
                        ('SSH Error: data could not be sent to the '
                         'remote host. Make sure this host can be '
                         'reached over ssh'))

            return (p.returncode, stdout, stderr)

    def _exec_command(self, cmd, in_data=None, sudoable=True):
        ''' run a command on the remote host '''

        super(Connection, self).exec_command(cmd, in_data=in_data,
                                             sudoable=sudoable)

        display.vvv((u"ESTABLISH SSH "
                     "CONNECTION FOR USER: {0}").format(
                         self._play_context.remote_user),
                    host=self._play_context.remote_addr)

        # we can only use tty when we are not pipelining the modules. piping
        # data into /usr/bin/python inside a tty automatically invokes the
        # python interactive-mode but the modules are not compatible with the
        # interactive-mode ("unexpected indent" mainly because of empty lines)

        if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
            ssh_executable = self._play_context.ssh_executable
        if in_data:
            if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
                cmd = self._build_command('ssh', self.host, cmd)
            if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
                cmd = self._build_command(ssh_executable, self.host, cmd)
        else:
            if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
                cmd = self._build_command('ssh', '-tt', self.host, cmd)
            if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
                cmd = self._build_command(ssh_executable, '-tt',
                                          self.host, cmd)

        (returncode, stdout, stderr) = self._run(cmd, in_data,
                                                 sudoable=sudoable)

        return (returncode, stdout, stderr)

    def dir_print(self, obj):
        for attr_name in dir(obj):
            try:
                attr_value = getattr(obj, attr_name)
                print(attr_name, attr_value, callable(attr_value))
            except Exception:
                pass

    def exec_command(self, cmd, in_data=None, sudoable=False):
        # Main public methods
        ''' run a command on the chroot '''
        display.vvv('XXX exec_command: %s' % cmd)
        super(Connection, self).exec_command(cmd, in_data=in_data,
                                             sudoable=sudoable)

        if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
            ssh_executable = self._play_context.ssh_executable
        # #print dir(self)
        # #print dir(self._play_context)
        # #print self._play_context._attributes
        # self.dir_print(self._play_context)
        # vm = self._play_context.get_ds()
        # print( vm )
        # raise "blah"
        h = self.container_name
        if (self.lxc_version == 2):
            lxc_cmd = 'lxc exec %s --mode=non-interactive -- /bin/sh -c %s'  \
                    % (pipes.quote(h),
                       pipes.quote(cmd))
        elif (self.lxc_version == 1):
            lxc_cmd = 'lxc-attach --name %s -- /bin/sh -c %s'  \
                    % (pipes.quote(h),
                       pipes.quote(cmd))
        if in_data:
            if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
                cmd = self._build_command('ssh', self.host, lxc_cmd)
            if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
                cmd = self._build_command(ssh_executable, self.host, lxc_cmd)
        else:
            if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
                cmd = self._build_command('ssh', '-tt', self.host, lxc_cmd)
            if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
                cmd = self._build_command(ssh_executable, '-tt',
                                          self.host, lxc_cmd)
        # self.ssh.exec_command(lxc_cmd,in_data,sudoable)
        (returncode, stdout, stderr) = self._run(cmd, in_data,
                                                 sudoable=sudoable)
        return (returncode, stdout, stderr)

    def put_file(self, in_path, out_path):
        ''' transfer a file from local to lxc '''
        super(Connection, self).put_file(in_path, out_path)
        if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
            display.vvv('XXX put_file %s %s' % (in_path, out_path))
        if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
            display.vvv(u"PUT {0} TO {1}".format(in_path, out_path),
                        host=self.host)
        ssh_executable = self._play_context.ssh_executable

        if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
            if not os.path.exists(in_path):
                raise AnsibleFileNotFound(("file or module does not "
                                           "exist: %s") % in_path)
        if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
            if not os.path.exists(to_bytes(
                                  in_path, errors='surrogate_or_strict')):
                raise AnsibleFileNotFound(
                        ("file or module does not "
                         "exist: {0}").format(to_native(in_path)))

        with open(in_path, 'rb') as in_f:
            in_data = in_f.read()
            if (len(in_data) == 0):
                # define a shortcut for empty files - nothing to read so the
                # ssh pipe will hang
                cmd = ('touch %s; echo -n done' % pipes.quote(out_path))
            else:
                # regular command
                cmd = ('cat > %s; echo -n done' % pipes.quote(out_path))
            h = self.container_name
            if (self.lxc_version == 2):
                lxc_cmd = ('lxc exec %s --mode=non-interactive -- '
                           '/bin/sh -c %s')  \
                        % (pipes.quote(h),
                           pipes.quote(cmd))
            elif (self.lxc_version == 1):
                lxc_cmd = 'lxc-attach --name %s -- /bin/sh -c %s'  \
                        % (pipes.quote(h),
                           pipes.quote(cmd))
            if in_data:
                if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
                    cmd = self._build_command('ssh', self.host, lxc_cmd)
                if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
                    cmd = self._build_command(ssh_executable, self.host,
                                              lxc_cmd)
            else:
                if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
                    cmd = self._build_command('ssh', '-tt', self.host, lxc_cmd)
                if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
                    cmd = self._build_command(ssh_executable, '-tt',
                                              self.host, lxc_cmd)
            # self.ssh.exec_command(lxc_cmd,in_data,sudoable)
            (returncode, stdout, stderr) = self._run(cmd, in_data,
                                                     sudoable=False)
            return (returncode, stdout, stderr)

    def fetch_file(self, in_path, out_path):
        ''' fetch a file from lxc to local '''
        super(Connection, self).fetch_file(in_path, out_path)
        if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
            display.vvv('XXX fetch_file %s %s' % (in_path, out_path))
        if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
            display.vvv(u"FETCH {0} TO {1}".format(in_path, out_path),
                        host=self.host)
        ssh_executable = self._play_context.ssh_executable

        if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
            cmd = ('cat %s' % pipes.quote(in_path))
        if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
            cmd = ('cat < %s' % pipes.quote(in_path))
        h = self.container_name
        if (self.lxc_version == 2):
            lxc_cmd = 'lxc exec %s --mode=non-interactive -- /bin/sh -c %s'  \
                    % (pipes.quote(h),
                       pipes.quote(cmd))
        elif (self.lxc_version == 1):
            lxc_cmd = 'lxc-attach --name %s -- /bin/sh -c %s'  \
                    % (pipes.quote(h),
                       pipes.quote(cmd))

        if LooseVersion(ansible_version) < LooseVersion('2.3.0.0'):
            in_data = None
            if in_data:
                cmd = self._build_command('ssh', self.host, lxc_cmd)
            else:
                cmd = self._build_command('ssh', '-tt', self.host, lxc_cmd)
            (returncode, stdout, stderr) = self._run(cmd, in_data,
                                                     sudoable=False)
            if returncode != 0:
                raise AnsibleError(("failed to transfer file "
                                    "from {0}:\n{1}\n{2}").format(in_path,
                                                                  stdout,
                                                                  stderr))
            with open(out_path, 'w') as out_f:
                out_f.write(stdout)

        if LooseVersion(ansible_version) >= LooseVersion('2.3.0.0'):
            cmd = self._build_command(ssh_executable, self.host, lxc_cmd)
            (returncode, stdout, stderr) = self._run(cmd, None, sudoable=False)

        if returncode != 0:
            raise AnsibleError(("failed to transfer file "
                                "from {0}:\n{1}\n{2}").format(in_path, stdout,
                                                              stderr))
        with open(out_path, 'w') as out_f:
            out_f.write(stdout)

        return (returncode, stdout, stderr)

    def reset(self):
        # only used from Ansible version 2.3 on forward
        # If we have a persistent ssh connection (ControlPersist), we can
        # ask it to stop listening.
        cmd = self._build_command(self._play_context.ssh_executable,
                                  '-O', 'stop', self.host)
        controlpersist, controlpath = self._persistence_controls(cmd)
        if controlpersist:
            display.vvv(u'sending stop: %s' % cmd)
            p = subprocess.Popen(cmd, stdin=subprocess.PIPE,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            stdout, stderr = p.communicate()
            status_code = p.wait()
            if status_code != 0:
                raise AnsibleError("Cannot reset connection:\n%s" % stderr)
        self.close()

    def close(self):
        ''' terminate the connection; nothing to do here '''
        display.vvv('XXX close')
        super(Connection, self).close()
        # self.ssh.close()
        self._connected = False



testinfra_hosts = ['ansible://debian-host']


def test_apt(host):
    with host.sudo():
        cmd = host.run("apt update")
        print(cmd.stdout)
        print(cmd.stderr)
        assert 0 == cmd.rc



#!/opt/venvs/securedrop-app-code/bin/python
# -*- coding: utf-8 -*-

import os
import sys

os.environ["SECUREDROP_ENV"] = "dev"  # noqa
import journalist_app

from sdconfig import config
from db import db
from models import Journalist


def main(username, password, otp_secret):
    app = journalist_app.create_app(config)
    with app.app_context():

        user = Journalist(username=username,
                          password=password,
                          is_admin=True,
                          first_name="",
                          last_name="")
        user.otp_secret = otp_secret
        db.session.add(user)
        db.session.commit()


main(*sys.argv[1:])



testinfra_hosts = ['ansible://bind-host', 'ansible://bind-client-host', 'ansible://icinga-host']


def test_sshfp(host):
    domain = host.run("hostname -d").stdout.strip()
    cmd = host.run("ssh -v "
                   "-o BatchMode=yes "
                   "-o VerifyHostKeyDNS=yes "
                   f"debian@bind-host.{domain} true || true")
    assert "debug1: matching host key fingerprint found in DNS" in cmd.stderr



import pytest
from django.urls import reverse
from rest_framework.authtoken.models import Token
from django.contrib.auth import get_user_model
from rest_framework.test import APITestCase
from rest_framework.test import APIClient
import mock


UserModel = get_user_model()


class APIUserAPITestCase(APITestCase):

    @pytest.mark.django_db
    def setUp(self):
        self.user = UserModel.objects.create_user(
            username='test', email='test@...', password='top_secret')
        token = Token.objects.create(user=self.user)
        self.client = APIClient()
        self.client.credentials(HTTP_AUTHORIZATION='Token ' + token.key)


class TestAPI(APIUserAPITestCase):

    @pytest.mark.django_db
    @mock.patch('enough.common.bind.delegate_dns')
    @mock.patch('enough.api.permissions.has_permission')
    def test_delegate_test_dns(self, has_permission, delegate_dns):
        url = reverse('delegate-test-dns')
        delegate_dns.return_value = [{"changed": False}, {"changed": True}, {"changed": False}]
        has_permission.return_value = True
        data = {
            "name": "gyztqojxhe3tinjrbi",
            "ip": "1.2.3.4",
        }

        response = self.client.post(url, data, format='json')

        assert has_permission.called
        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.json()[0], {"changed": False})

    @pytest.mark.django_db
    @mock.patch('enough.common.bind.delegate_dns')
    def test_delegate_dns(self, delegate_dns):
        url = reverse('delegate-dns')
        delegate_dns.return_value = [{"changed": False}, {"changed": True}, {"changed": False}]
        data = {
            "name": "gyztqojxhe3tinjrbi",
            "ip": "1.2.3.4",
        }

        response = self.client.post(url, data, format='json')

        self.assertEqual(response.status_code, 201)
        self.assertEqual(response.json()[0], {"changed": False})

    @pytest.mark.django_db
    def test_ping(self):
        url = reverse('ping')
        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b'pong')



testinfra_hosts = ['ansible://icinga-host']


def test_ns1_test(host):
    domain = host.run("hostname -d").stdout.strip()
    cmd = host.run("dig +short NS test.{}".format(domain))
    assert 0 == cmd.rc
    assert "ns1.{}.".format(domain) == cmd.stdout.strip()


def test_bind(host):
    domain = host.run("hostname -d").stdout.strip()
    cmd = host.run("getent hosts ns1.{}".format(domain))
    assert 0 == cmd.rc


def test_update(host):
    domain = host.run("hostname -d").stdout.strip()
    hostname = host.run("hostname -s").stdout.strip()
    bind_host = host.get_host('ansible://bind-host',
                              ansible_inventory=host.backend.ansible_inventory)
    address = bind_host.ansible.get_variables()['ansible_host']
    cmd = bind_host.run(f'''
        nsupdate <<EOF
        server {address}
        zone test.{domain}
        update add {hostname}.test.{domain}. 1800 TXT "Updated by nsupdate"
        show
        send
        quit
        EOF
        ''')
    assert 0 == cmd.rc


def test_dig_update(host):
    domain = host.run("hostname -d").stdout.strip()
    hostname = host.run("hostname -s").stdout.strip()
    cmd = host.run("dig +short TXT {}.test.{}".format(hostname, domain))
    print(cmd.stdout.strip())
    assert 0 == cmd.rc
    assert "Updated by nsupdate" in cmd.stdout.strip()


def test_clean_update(host):
    domain = host.run("hostname -d").stdout.strip()
    hostname = host.run("hostname -s").stdout.strip()
    host = host.get_host('ansible://bind-host',
                         ansible_inventory=host.backend.ansible_inventory)
    address = host.ansible.get_variables()['ansible_host']
    cmd = host.run(f'''
        nsupdate <<EOF
        server {address}
        zone test.{domain}
        update delete {hostname}.test.{domain}. TXT
        show
        send
        quit
        EOF
        ''')
    assert 0 == cmd.rc



import json
import logging

testinfra_hosts = ['ansible://infrastructure5-host']
logger = logging.getLogger(__name__)


def test_no_volume(host):
    logger = logging.getLogger(__name__)
    cmd = host.run('lsblk --list --json')
    logger.debug("stdout: %s", cmd.stdout)
    logger.debug("stderr: %s", cmd.stderr)
    blockdevices = json.loads(cmd.stdout)['blockdevices']
    # check that there is only one disk device
    assert 1 == len([disk for disk in blockdevices if disk['type'] == 'disk'])
    # check that there is only one partition (root)
    assert 1 == len([disk for disk in blockdevices
                     if disk['type'] == 'part' and disk['mountpoint'] == '/'])
    assert 0 == cmd.rc



import logging
import os
import re
import sys

from cliff.app import App
from cliff.commandmanager import CommandManager

from enough.version import __version__

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'enough.settings')


class EnoughApp(App):

    def __init__(self, modules):
        super().__init__(
            description='enough',
            version=__version__,
            command_manager=CommandManager(modules),
            deferred_help=True,
            )

    @staticmethod
    def set_enough_domain(argv):
        capture = False
        for a in argv:
            if capture:
                os.environ['ENOUGH_DOMAIN'] = a
                break
            if a == '--domain':
                capture = True

    def configure_logging(self):
        super().configure_logging()

        root_logger = logging.getLogger('')
        root_logger.setLevel(logging.WARNING)

        level = {0: logging.WARNING,
                 1: logging.INFO,
                 2: logging.DEBUG}.get(self.options.verbose_level, logging.DEBUG)

        root_logger = logging.getLogger('enough')
        root_logger.setLevel(level)

    def build_option_parser(self, description, version, argparse_kwargs=None):
        parser = super().build_option_parser(description, version, argparse_kwargs)
        parser.add_argument('--domain', default='enough.community', help='Enough domain name')
        return parser

    @staticmethod
    def remap(argv):
        home = os.path.expanduser('~/.enough')

        def remap(s):
            s = re.sub(r'^/\S+/.enough', home, s)
            s = re.sub(r'=/\S+/.enough', f'={home}', s)
            return s
        return [remap(a) for a in argv]

    @staticmethod
    def preserve_ownership():
        uid = os.geteuid()
        home = os.path.expanduser('~/.enough')
        if not uid == 0 or not os.path.exists(home):
            return False
        home_uid = os.stat(home).st_uid
        if uid != home_uid:
            os.system(f'chown -R {home_uid} {home}')
        return True

    def run(self, argv):
        r = super().run(self.remap(argv))
        self.preserve_ownership()
        return r


def main(argv=sys.argv[1:]):
    EnoughApp.set_enough_domain(argv)
    myapp = EnoughApp('enough.cli')
    return myapp.run(argv)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



import logging

testinfra_hosts = ['ansible://infrastructure2-host']
logger = logging.getLogger(__name__)


def test_infrastructure2_encrypted_volume(host):
    cmd = host.run("""
    set -xe
    test -e /dev/mapper/spare
    grep -q /opt /etc/fstab
    test -e /opt/docker/image
    test -e /opt/snap/hello-world
    grep -q snap /etc/fstab
    """)
    logger.debug('stdout %s', cmd.stdout)
    logger.debug('stderr %s', cmd.stderr)
    assert 0 == cmd.rc

    # set by infrastructure/test-encrypted-volume.yml
    with host.sudo():
        host.file("/etc/cryptsetup/keyfile").content == "PASSWORD"



testinfra_hosts = ['ansible://bind-host']


def openstack(host, cmd):
    cmd = host.run(f"""
    export OS_CLIENT_CONFIG_FILE=/usr/lib/backup/clouds.yml
    export OS_CLOUD=production
    openstack {cmd}
    """)
    print(cmd.stderr)
    assert 0 == cmd.rc
    return cmd.stdout


def expected_snapshots(host, count):
    assert count == openstack(
        host, "volume snapshot list -f value -c Name | grep -c pet-volume").strip()


def test_snapshots(host):
    with host.sudo():
        cmd = host.run("/etc/cron.daily/snapshot")
        print(cmd.stdout)
        print(cmd.stderr)
        assert 0 == cmd.rc

        expected_snapshots(host, '1')
        cmd = host.run("/etc/cron.daily/snapshot 0")
        print(cmd.stdout)
        print(cmd.stderr)
        assert 0 == cmd.rc

        assert openstack(host, "volume snapshot list -f value -c Name") == ""



testinfra_hosts = ['ansible://client-host']


def test_server_visible_from_client(host):
    server_host = host.get_host('ansible://server-host',
                                ansible_inventory=host.backend.ansible_inventory)
    server_ip = server_host.ansible.get_variables()['ansible_host']
    with host.sudo():
        host.run("apt-get install -y nmap")

    #
    # TCP
    #
    r = host.run("nmap  -oG - -Pn -p 1-1024,5665 " + server_ip)
    assert r.rc == 0
    assert ("Ports: "
            "22/open/tcp//ssh///, "
            "53/closed/tcp//domain///, "
            "80/closed/tcp//http///, "
            "443/closed/tcp//https///, "
            "465/closed/tcp//smtps///, "
            "5665/closed/tcp//unknown///\t") in r.stdout
    #
    # UDP
    #
    with host.sudo():
        r = host.run("nmap  -oG - -sU -Pn " + server_ip)
        assert r.rc == 0
    assert ("Ports: "
            "53/closed/udp//domain///\t") in r.stdout
    #
    # ICMP
    #
    r = host.run("ping -W 5 -c 1 " + server_ip)
    assert r.rc == 0
    assert ', 0% packet loss' in r.stdout


def test_server_visible_from_external(host):
    external_host = host.get_host('ansible://external-host',
                                  ansible_inventory=host.backend.ansible_inventory)
    server_host = host.get_host('ansible://server-host',
                                ansible_inventory=host.backend.ansible_inventory)
    server_ip = server_host.ansible.get_variables()['ansible_host']
    with external_host.sudo():
        external_host.run("apt-get install -y nmap")
    #
    # TCP
    #
    r = external_host.run("nmap  -oG - -Pn " + server_ip)
    assert r.rc == 0
    assert ("Ports: "
            "22/open/tcp//ssh///, "
            "53/closed/tcp//domain///, "
            "80/closed/tcp//http///, "
            "443/closed/tcp//https///\t") in r.stdout
    #
    # ICMP
    #
    r = external_host.run("ping -W 5 -c 1 " + server_ip)
    assert r.rc == 1
    assert ', 100% packet loss' in r.stdout


def test_gitlab_visible_from_client(host):
    gitlab_host = host.get_host('ansible://gitlab-host',
                                ansible_inventory=host.backend.ansible_inventory)
    gitlab_ip = gitlab_host.ansible.get_variables()['ansible_host']
    with host.sudo():
        host.run("apt-get install -y nmap")
    #
    # TCP
    #
    r = host.run("nmap  -oG - -Pn -p 2222 " + gitlab_ip)
    assert r.rc == 0
    assert ("Ports: "
            "2222/open/tcp//EtherNetIP-1///") in r.stdout



testinfra_hosts = ['ansible://bind-host']


def expected_backups(host, count):
    cmd = host.run("""
    export OS_CLIENT_CONFIG_FILE=/usr/lib/backup/clouds.yml
    export OS_CLOUD=production
    openstack ${OS_INSECURE} image list | grep -c pet-host
    """)
    print(cmd.stderr)
    assert count == cmd.stdout.strip()
    assert 0 == cmd.rc


def test_backup(host):
    # we need --insecure during tests otherwise going back in time a few days
    # may invalidate some certificates and result in errors such as:
    # SSL exception connecting to
    #    https://auth.cloud.ovh.net/v2.0/tokens: [SSL: CERTIFICATE_VERIFY_FAILED]
    with host.sudo():
        cmd = host.run("OS_INSECURE=--insecure /etc/cron.daily/prune-backup 0")
        print(cmd.stdout)
        print(cmd.stderr)
        assert 0 == cmd.rc
        try:
            host.run("timedatectl set-ntp 0")
            cmd = host.run("""
            set -x
            export OS_INSECURE=--insecure
            date -s '-15 days'
            bash -x /etc/cron.daily/backup
            date -s '-30 days'
            bash -x /etc/cron.daily/backup
            """)
            host.run("timedatectl set-ntp 1")
            print(cmd.stdout)
            print(cmd.stderr)
            assert 0 == cmd.rc
            expected_backups(host, '2')
            host.run("OS_INSECURE=--insecure bash -x /etc/cron.daily/prune-backup 30")
            expected_backups(host, '1')
        finally:
            host.run("timedatectl set-ntp 1")
            host.run("OS_INSECURE=--insecure bash -x /etc/cron.daily/prune-backup 0")



import os
import pytest
import shutil
import yaml

from enough.common import Enough
from tests import make_config_dir, prepare_config_dir


def pytest_addoption(parser):
    parser.addoption(
        "--enough-no-create",
        action="store_true",
        help="Do not run the create step"
    )
    parser.addoption(
        "--enough-no-tests",
        action="store_true",
        help="Do not run the tests step"
    )
    parser.addoption(
        "--enough-no-destroy",
        action="store_true",
        help="Do not run the destroy step"
    )


def pytest_configure(config):
    pass


def get_hosts(session, enough):
    hosts = session.config.getoption("--enough-hosts").split(',')
    updated = enough.service.add_vpn_hosts_if_needed(hosts)
    return (enough.service.get_vpn_host(), updated)


def pytest_sessionstart(session):
    if session.config.getoption("--enough-no-create"):
        return

    if not session.config.getoption("--enough-no-destroy"):
        enough_destroy(session)

    service_directory = session.config.getoption("--enough-service")
    domain = f'{service_directory}.test'
    enough_dot_dir = session.config.cache.makedir('dotenough')
    config_dir = prepare_config_dir(domain, enough_dot_dir)

    e = Enough(config_dir, '.',
               domain=domain,
               driver='openstack',
               inventory=[f'playbooks/{service_directory}/inventory'],
               route_to_internal=False)
    (vpn_host, hosts) = get_hosts(session, e)
    r = e.create_missings(hosts)
    with open(f'{e.config_dir}/inventory/group_vars/all/domain.yml', 'r') as f:
        data = yaml.safe_load(f)
    if len(r) > 0 or data.get('domain') == domain:
        e.heat.create_test_subdomain('enough.community')
    if vpn_host in hosts:
        if not e.vpn_has_credentials():
            bind_host = e.service.service2group.get('bind')[0]
            e.playbook.run([
                '--private-key', f'{e.config_dir}/infrastructure_key',
                '-i', f'playbooks/openvpn/inventory',
                f'--limit={vpn_host},{bind_host},localhost',
                'playbooks/openvpn/conftest-playbook.yml',
            ])
        e.vpn_connect()

    hosts = ','.join(hosts)
    e.playbook.run([
        '--private-key', f'{e.config_dir}/infrastructure_key',
        '-i', f'playbooks/{service_directory}/inventory',
        f'--limit={hosts},localhost',
        f'playbooks/{service_directory}/playbook.yml',
    ])
    if vpn_host and vpn_host in hosts:
        e.vpn_disconnect()


def pytest_runtest_setup(item):
    if item.config.getoption("--enough-no-tests"):
        pytest.skip("--enough-no-tests specified, skipping all tests")


def pytest_sessionfinish(session, exitstatus):
    if not session.config.getoption("--enough-no-destroy"):
        enough_destroy(session)


def enough_destroy(session):
    service_directory = session.config.getoption("--enough-service")
    domain = f'{service_directory}.test'
    enough_dot_dir = session.config.cache.makedir('dotenough')
    config_dir = make_config_dir(domain, enough_dot_dir)
    all_dir = f'{config_dir}/inventory/group_vars/all'
    if not os.path.exists(all_dir):
        os.makedirs(all_dir)
    shutil.copyfile('inventory/group_vars/all/clouds.yml', f'{all_dir}/clouds.yml')
    e = Enough(config_dir, '.',
               domain=domain,
               driver='openstack',
               inventory=[f'playbooks/{service_directory}/inventory'])
    e.destroy()


def pytest_unconfigure(config):
    pass



#!/usr/bin/env python

# cran.py: install or remove R packages
# Homepage: https://github.com/yutannihilation/ansible-module-cran

# Copyright (C) 2016 Hiroaki Yutani <yutani.ini@gmail.com>
# Copyright (C) 2017 Maciej Delmanowski <drybjed@gmail.com>
# Copyright (C) 2017 DebOps https://debops.org/

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from ansible.module_utils.basic import AnsibleModule

DOCUMENTATION = '''
---
module: cran
short_description: Install R packages.
options:
  name:
    description:
      - The name of an R package.
    required: true
    default null
  state:
    description:
      - The state of module
    required: false
    choices: ['present', 'absent']
    default: present
  repo:
    description:
      - The repository
    required: false
    default: "https://cran.rstudio.com/"
'''

RSCRIPT = '/usr/bin/Rscript'


def get_installed_version(module):
    cmnd = [RSCRIPT, '--slave', '--no-save', '--no-restore-history', '-e',
            'p <- installed.packages(); cat(p[p[,1] == "{name:}",'
            '3])'.format(name=module.params['name'])]
    (rc, stdout, stderr) = module.run_command(cmnd, check_rc=False)
    return stdout.strip() if rc == 0 else None


def install(module):
    cmnd = [RSCRIPT, '--slave', '--no-save', '--no-restore-history', '-e',
            'install.packages(pkgs="{name:}",repos="{repos:}")'
            ''.format(name=module.params['name'],
                      repos=module.params['repo'])]
    (rc, stdout, stderr) = module.run_command(cmnd, check_rc=True)
    return stderr


def uninstall(module):
    cmnd = [RSCRIPT, '--slave', '--no-save', '--no-restore-history', '-e',
            'remove.packages(pkgs="{name:}")'
            ''.format(name=module.params['name'])]
    (rc, stdout, stderr) = module.run_command(cmnd, check_rc=True)
    return stderr


def main():
    module = AnsibleModule(
        argument_spec=dict(
            state=dict(default='present', choices=['present', 'absent']),
            name=dict(required=True),
            repo=dict(default='https://cran.rstudio.com/')
        )
    )
    state = module.params['state']
    name = module.params['name']
    changed = False
    version = get_installed_version(module)

    if state == 'present' and not version:
        stderr = install(module)
        version = get_installed_version(module)
        if not version:
            module.fail_json(
                msg='Failed to install {name:}: {err:}'.format(
                    name=name, err=stderr, version=version))
        changed = True

    elif state == 'absent' and version:
        stderr = uninstall(module)
        version = get_installed_version(module)
        if version:
            module.fail_json(
                msg='Failed to install {name:}: {err:}'.format(
                    name=name, err=stderr))
        changed = True

    module.exit_json(changed=changed, name=name, version=version)


if __name__ == '__main__':
    main()



from cliff.command import Command

from enough import settings
from enough.common import options
from enough.common.openstack import OpenStack


class Cli(Command):
    "OpenStack client"

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument('args', nargs='+')
        return options.set_options(parser)

    def take_action(self, parsed_args):
        args = vars(self.app.options)
        args.update(vars(parsed_args))
        o = OpenStack(settings.CONFIG_DIR, **args)
        o.run(args['args'])



import os
import sh

from cliff.command import Command

from enough.common.docker import Docker


class Build(Command):
    "Build the enough docker image."

    class DockerEnough(Docker):

        def create_image(self):
            name = super().create_image()
            sh.rm('-fr', 'dist')
            sh.python('setup.py', '--quiet', 'sdist')
            dockerfile = os.path.join(self.root, 'internal/data/enough-source.dockerfile')
            return self._create_image(None,
                                      '--build-arg', f'IMAGE_NAME={name}',
                                      '-f', dockerfile, '.')

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument('--name', default='enough')
        return parser

    def take_action(self, parsed_args):
        args = vars(parsed_args)
        args['domain'] = self.app.options.domain
        Build.DockerEnough(**args).create_image()


class Create(Command):
    "Create the enough service."

    class DockerEnough(Docker):

        def get_compose_content(self):
            f = os.path.join(self.root, 'internal/data/enough-docker-compose.yml')
            return self.replace_content(open(f).read())

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument('--name', default='enough')
        parser.add_argument('--port', default='8000')
        return parser

    def take_action(self, parsed_args):
        args = vars(parsed_args)
        args['domain'] = self.app.options.domain
        Create.DockerEnough(**args).up_wait_for_services()



from tests.icinga_helper import IcingaHelper

testinfra_hosts = ['ansible://icinga-host']


class TestChecks(IcingaHelper):

    def test_host(self):
        assert 'website-host' in self.get_hosts(host='website-host')

    def test_service_disk(self):
        assert self.is_service_ok('website-host!disk')

    def test_service_openvpn(self):
        assert self.is_service_ok('website-host!Check OpenVPN')



# -*- coding: utf-8 -*-

# Copyright (C) 2014-2015 Hartmut Goebel <h.goebel@crazy-compilers.com>
# Part of the DebOps - https://debops.org/

# This program is free software; you can redistribute
# it and/or modify it under the terms of the
# GNU General Public License as published by the Free
# Software Foundation; either version 3 of the License,
# or (at your option) any later version.
#
# This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General
# Public License along with this program; if not,
# write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# An on-line copy of the GNU General Public License can
# be downloaded from the FSF web page at:
# https://www.gnu.org/copyleft/gpl.html

from __future__ import print_function

import sys
import os
import subprocess
import stat
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
try:
    # shlex.quote is new in Python 3.3
    from shlex import quote as shquote
except ImportError:
    # implement subset of shlex.quote
    def shquote(s):
        if not s:
            return "''"
        return "'" + s.replace("'", "'\"'\"'") + "'"

from .config import *

__author__ = "Hartmut Goebel <h.goebel@crazy-compilers.com>"
__copyright__ = "Copyright 2014-2015 by Hartmut Goebel "
"<h.goebel@crazy-compilers.com>"
__licence__ = "GNU General Public License version 3 (GPL v3) or later"


# ---- Global constants ----

ANSIBLE_CONFIG_FILE = "ansible.cfg"

# Path to the Ansible playbooks and roles distributed inside of the Python
# package
DEBOPS_PY_PACKAGE = os.path.join(os.path.dirname(__file__), 'ansible')

# --- Roles

# Default role prefix if no roles with prefixes are specified
ROLE_PREFIX = "debops"

# --- Playbooks

# Default site.yml playbook to look for
DEBOPS_SITE_PLAYBOOK = os.path.join("playbooks", "site.yml")

# --- Inventories

INVENTORY = "inventory"

# List of possible inventory directories, relative to DebOps root
# project directory
ANSIBLE_INVENTORY_PATHS = [
    os.path.join("ansible", INVENTORY),
    INVENTORY]

# --- Encryption

# Name of the script used to unlock/lock the encrypted directory
PADLOCK_CMD = "padlock"

ENCFS_CONFIGFILE = ".encfs6.xml"
ENCFS_PREFIX = ".encfs."
SECRET_NAME = "secret"
# Name of the keyfile stored inside EncFS encrypted directory
ENCFS_KEYFILE = ".encfs6.keyfile"
# Length of the random EncFS password stored in encrypted keyfile
ENCFS_KEYFILE_LENGTH = 256

# External programms used. List here for easy substitution for
# hard-coded paths.
ENCFS = 'encfs'
FUSERMOUNT = 'fusermount'
UMOUNT = 'umount'
GPG = 'gpg'


# ---- Functions ----

def _find_up(path, name):
    """
    Find specified file or directory in parent dir
    """
    # :todo: only walk up to the mount-point like git does
    path = os.path.abspath(path)
    last = None
    while path != last:
        last = path
        path = os.path.join(path, name)
        if os.path.exists(path):
            return path
        path = os.path.dirname(last)
    return None


def find_debops_project(path=None):
    """
    Searches DebOps, this workdir belongs to, if any.

    `path` defaults to the current working directory.

    Returns None if this path does not belong to a DebOps project.
    """
    if path is None:
        path = os.getcwd()
    # Find DebOps configuration file
    debops_config = _find_up(path, DEBOPS_CONFIG)
    # Find root of the DebOps project dir
    return os.path.dirname(debops_config) if debops_config else None


def find_monorepopath(config, project_root):
    """
    Search for monorepo in various locations.
    """
    if project_root:
        places = [os.path.join(project_root, "monorepo")]
    else:
        places = []
    places.append(config['paths']['monorepo-path'])
    for monorepo_path in places:
        if os.path.exists(os.path.join(monorepo_path, "LICENSE")):
            return monorepo_path


def find_playbookpath(config, project_root):
    """
    Search for playbooks in various locations.
    """
    if project_root:
        places = [os.path.join(project_root, "ansible", "playbooks")]
    else:
        places = []
    places.extend(config['paths']['playbooks-paths'])
    for playbook_path in places:
        if os.path.exists(os.path.join(playbook_path, "site.yml")):
            return playbook_path


def find_inventorypath(project_root):
    """
    Search Ansible inventory in local directories.
    """
    for inventory_path in ANSIBLE_INVENTORY_PATHS:
        ansible_inventory = os.path.join(project_root, inventory_path)
        if os.path.isdir(ansible_inventory):
            return ansible_inventory


# ---- Encryption support ----

def padlock_lock(encrypted_path):
    """
    Lock the padlock (this is: unmount the directory).

    Returns True if the padlock originally was unlocked, otherwise False.
    """
    # Cut the EncFS directory prefix to get the decrypted directory name
    decrypted_path = ''.join(encrypted_path.rsplit(ENCFS_PREFIX, 1))
    if not os.path.ismount(decrypted_path):
        return False
    # OS X compatibility
    if sys.platform == 'darwin':
        subprocess.call([UMOUNT, decrypted_path])
    else:
        subprocess.call([FUSERMOUNT, '-u', decrypted_path])
    return True


def padlock_unlock(encrypted_path):
    """
    Unlock the padlock (this is: mount the directory).

    Returns True if the padlock originally was locked, otherwise False.
    """
    # Location of GPG-encrypted keyfile to use
    keyfile = os.path.join(encrypted_path, ENCFS_KEYFILE)
    configfile = os.path.join(encrypted_path, ENCFS_CONFIGFILE)
    crypted_configfile = configfile+'.asc'

    # Location of an alternative executable that will mount the decrypted dir
    unlock_cmd = os.path.join(encrypted_path, PADLOCK_CMD)

    # Cut the EncFS directory prefix to get the decrypted directory name
    decrypted_path = ''.join(encrypted_path.rsplit(ENCFS_PREFIX, 1))

    if os.path.exists(keyfile) and os.path.exists(crypted_configfile):
        print("Mounting '{}' using encfs..".format(encrypted_path))
        unlock_cmd = None
    elif os.access(unlock_cmd, os.X_OK):
        print(
            "Mounting '{}' using '{}'..".format(encrypted_path, unlock_cmd))
    else:
        return False

    # Check if encrypted directory is already mounted
    # NB: This only tests if encfs_decrypted as a mount-point at all,
    # no matter if it is fuse.encfs or not.
    if os.path.ismount(decrypted_path):
        # if it is already mounted, do nothing
        return False

    # Make sure the mount directory exists
    if not os.path.isdir(decrypted_path):
        os.makedirs(decrypted_path)

    if unlock_cmd is not None:
        subprocess.check_output([unlock_cmd, encrypted_path, decrypted_path],
                                stderr=subprocess.STDOUT)
        return True

    # Make sure the named pipe for the configfile exists
    if not os.path.exists(configfile):
        os.mkfifo(configfile)
    elif not stat.S_ISFIFO(os.stat(configfile).st_mode):
        raise IOError(17, configfile+' exists but is not a fifo')

    # Start encfs. It will wait for input on the `configfile` named
    # pipe.
    encfs = subprocess.Popen([
        ENCFS, encrypted_path, decrypted_path,
        '--extpass',
        GPG + ' --decrypt --no-mdc-warning --output - %s' % shquote(keyfile)])
    # now decrypt the config and write it into the named pipe
    with open(configfile, 'w') as fh:
        # NB: gpg must write to stdout to avoid it is asking whether
        # the file should be overwritten
        subprocess.Popen([GPG,
                          '--decrypt', '--no-mdc-warning', '--output', '-',
                          crypted_configfile], stdout=fh).wait()
    encfs.wait()
    os.remove(configfile)
    return True



import logging
from enough.common import retry

testinfra_hosts = ['ansible://psono-host']
logger = logging.getLogger(__name__)


def test_psono_send_mail(host):

    with host.sudo():
        host.run("postsuper -d ALL")

    cmd = host.run("""
    cd /srv/psono
    sudo docker-compose run server python3 \
        ./psono/manage.py sendtestemail loic-doomtofail@dachary.org
    """)
    logger.debug('stdout %s', cmd.stdout)
    logger.debug('stderr %s', cmd.stderr)
    assert 0 == cmd.rc

    @retry.retry(AssertionError, tries=5)
    def wait_for_mail():
        with host.sudo():
            cmd = host.run("""
            grep -q doomtofail /var/spool/postfix/hold/*
            """)
        assert cmd.rc == 0, f'{cmd.stdout} {cmd.stderr}'
    wait_for_mail()



import os
from enough.common import docker
import pytest
from tests.modified_environ import modified_environ
from enough.version import __version__
from enough.common import retry


class DockerFixture(docker.Docker):

    def __init__(self, *args, **kwargs):
        # catch unintended calls to docker
        kwargs.setdefault('docker', '/bin/false')
        super(DockerFixture, self).__init__(*args, **kwargs)


def test_init():
    name = 'NAME'
    d = DockerFixture(name=name)
    assert d.name == name


def test_get_image_name():
    docker_name = 'DOCKER_NAME'
    suffix = 'SUFFIX'
    d = DockerFixture(name=docker_name)
    assert d.get_image_name(suffix) == docker_name + '_' + suffix
    assert d.get_image_name(None) == docker_name


def test_get_image_name_with_version():
    docker_name = 'DOCKER_NAME'
    suffix = 'SUFFIX'
    with modified_environ('DOCKER_HOST'):
        d = DockerFixture(name=docker_name)
        image_name = d.get_image_name(suffix) + ':' + str(__version__)
        assert d.get_image_name_with_version(suffix) == image_name


def test_replace_content():
    docker_name = 'DOCKER_NAME'
    with modified_environ('DOCKER_HOST'):
        d = DockerFixture(name=docker_name)
        before = 'Name = {{ this.get_image_name_with_version(None) }}'
        after = 'Name = ' + d.get_image_name_with_version(None)
        assert d.replace_content(before) == after


def test_up_no_wait(docker_name, tcp_port):
    d = docker.Docker(name=docker_name, port=tcp_port)
    assert d.create_image()
    d.up()
    assert d.inspect('{{ .Path }}') == ['/sbin/init']
    d.down()


def test_up_wait_for_services(docker_name, tcp_port):
    d = docker.Docker(name=docker_name, port=tcp_port)
    assert d.create_image()
    d.up_wait_for_services()
    assert '"Status":"healthy"' in d.get_logs()
    assert d.docker_compose.exec(
        '-T', docker_name, 'hostname').stdout.strip().decode('utf-8') == docker_name
    d.down()


def test_create_or_update(docker_name, tcp_port):
    d = docker.Docker(name=docker_name, port=tcp_port)
    assert d.create_or_update() is True
    assert d.create_or_update() is False


def test_get_public_port(docker_name, tcp_port):
    d = docker.Docker(name=docker_name, port=tcp_port)
    assert d.get_public_port('8000') is None
    assert d.create_image()
    d.up_wait_for_services()
    assert d.get_public_port('8000') == tcp_port
    d.down()


def test_get_ip(docker_name, tcp_port):
    d = docker.Docker(name=docker_name, port=tcp_port)
    assert d.create_image()
    d.up_wait_for_services()
    assert '.' in d.get_ip()
    d.down()


def test_up_wait_for_services_fail(docker_name):
    class DockerFixtureIntegration(docker.Docker):

        def __init__(self, *args, **kwargs):
            kwargs['root'] = 'tests/enough/common/data'
            kwargs['retry'] = 2
            super().__init__(*args, **kwargs)

        def get_compose_content(self):
            f = os.path.join(self.root, 'common/data/docker-compose-fail.yml')
            return self.replace_content(open(f).read())

    d = DockerFixtureIntegration(name=docker_name)
    assert d.create_image()
    with pytest.raises(retry.RetryException):
        d.up_wait_for_services()
    d.down()


def test_create_image(docker_name):
    d = docker.Docker(name=docker_name)
    assert d.create_image().startswith(docker_name)


def test_create_network(docker_name):
    d = docker.Docker(name=docker_name)
    assert d.create_network(docker_name) is True
    assert d.create_network(docker_name) is False



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016-2018 DebOps project https://debops.org/

from __future__ import print_function
import os
import fnmatch
import re
from subprocess import check_output


# Fix "Edit on GitHub" links in the documentation
# Jinja2 Support is only basic Jinja2 without all the good stuff from Ansible.
# So I am not gonna mess with that or try to extend it as in:
# https://stackoverflow.com/questions/36019670/
# What I am gonna do instead is just recompute source file to URL map in Python
# and job done.

def find_files(directory, pattern):
    for root, dirs, files in os.walk(directory):
        for basename in files:
            if fnmatch.fnmatch(basename, pattern):
                filename = os.path.join(root, basename)
                yield filename


def get_source_file_to_url_map(start_dir='.', skip_patterns=[]):
    source_file_to_url_map = {}
    repo_dir_to_url_map = {}

    cur_dir = os.path.abspath(start_dir)

    for source_file_name in find_files('.', '*.rst'):
        pagename_source_file = source_file_name.lstrip('/.')

        skip = False
        for skip_pattern in skip_patterns:
            if re.search(skip_pattern, pagename_source_file):
                skip = True
                break

        if skip:
            continue

        dir_path = os.path.dirname(source_file_name)
        if len(dir_path) > 2:
            dir_path = dir_path.lstrip('/.')

        # Can also contain subdirs in a repo but this optimization should
        # already get factor 10 in performance for git invocation.
        if dir_path not in repo_dir_to_url_map:
            for remote_line in check_output(['git', '-C', dir_path,
                                             'remote', '-v']).split(b'\n'):
                remote_item = re.split(r'\s', remote_line.decode('utf-8'))
                if remote_item[0] == 'origin' and remote_item[2] == '(fetch)':
                    base_url = remote_item[1]

                    if base_url.endswith('.git'):
                        base_url = base_url[:-4]

                    if '@' in base_url:  # it's a git+ssh URL
                        base_url = base_url.replace(':', '/')
                        base_url = base_url.replace('git@', 'https://')
                    repo_dir_to_url_map[dir_path] = base_url

        relative_pagename = 'docs/' + pagename_source_file

        if re.match(r'docs/news/changelog(?:\.rst)$',
                    relative_pagename, flags=re.I):
            relative_pagename = 'CHANGELOG.rst'

        if re.match(r'docs/user-guide/install(?:\.rst)$',
                    relative_pagename, flags=re.I):
            relative_pagename = 'INSTALL.rst'

        relative_path = re.match(
                r'docs/ansible/roles/(.*)/defaults/(.+)\.rst$',
                relative_pagename, flags=re.I)
        if relative_path:
            relative_pagename = ('ansible/roles/' + relative_path.group(1)
                                 + '/defaults/' + relative_path.group(2)
                                 + '.yml')

        pagename_source_file = re.sub(r'\.rst$', '', pagename_source_file)
        source_file_to_url_map[pagename_source_file] = {
            'url': repo_dir_to_url_map[dir_path],
            'pagename': relative_pagename,
        }

    return source_file_to_url_map



import os
import pytest
import requests_mock as requests_mock_module

from enough import settings
from enough.common import openstack
from enough.common import service

from tests import prepare_config_dir


@pytest.mark.openstack_integration
def test_openstack_create_or_update(request, tmpdir, requests_mock):
    try:
        domain = 'enough.test'
        config_dir = prepare_config_dir(domain, tmpdir)
        requests_mock.post(requests_mock_module.ANY, status_code=201)
        requests_mock.get(requests_mock_module.ANY, status_code=200)
        playbook = 'tests/enough/common/test_common_service/enough-playbook.yml'
        s = service.ServiceOpenStack(config_dir, settings.SHARE_DIR, **{
            'driver': 'openstack',
            'playbook': os.path.abspath(playbook),
            'domain': domain,
            'name': 'essential',
            'cloud': 'production',
        })
        s.dotenough.set_certificate('ownca')
        s.dotenough.set_clouds_file('inventory/group_vars/all/clouds.yml')
        r = s.create_or_update()
        assert r['fqdn'] == f'essential.{domain}'
        # the second time around the hosts.yml are reused
        r = s.create_or_update()
        assert r['fqdn'] == f'essential.{domain}'

    finally:
        o = openstack.OpenStack(s.config_dir, **s.args)
        # comment out the following line to re-use the content of the regions and save time
        o.destroy_everything(None)


def test_service_from_host(tmpdir):
    s = service.Service(config_dir=tmpdir, share_dir=settings.SHARE_DIR,
                        domain='test.com')
    assert s.service_from_host('icinga-host') is None
    assert s.service_from_host('cloud-host') == 'cloud'
    assert s.service_from_host('unknown-host') is None


def test_set_service_info(tmpdir):
    s = service.Service(config_dir=tmpdir, share_dir=settings.SHARE_DIR,
                        domain='test.com')
    assert 'bind-host' in s.service2hosts['bind']
    assert len(s.service2hosts['bind']) > 0
    assert ['bind-host'] == s.service2group['bind']


def test_update_vpn_dependencies(tmpdir):
    s = service.Service(config_dir=tmpdir, share_dir=settings.SHARE_DIR,
                        domain='test.com')
    assert s.hosts_with_internal_network(['bind-host']) == []
    assert 'website-host' not in s.service2hosts['openvpn']
    assert 'website-host' not in s.service2hosts['weblate']
    inventories = ('tests/enough/common/test_common_service/vpn_inventory',)
    s = service.Service(config_dir=tmpdir, share_dir=settings.SHARE_DIR,
                        domain='test.com', inventory=inventories)
    assert 'website-host' in s.service2hosts['openvpn']
    assert s.hosts_with_internal_network(['icinga-host']) == ['icinga-host']
    s.update_vpn_dependencies()
    assert 'website-host' not in s.service2hosts['weblate']
    assert 'weblate-host' in s.service2hosts['weblate']


def test_ensure_non_empty_service_group(tmpdir):
    name = 'wekan'
    s = service.Service(tmpdir, settings.SHARE_DIR,
                        name=name,
                        domain='test.com')
    with pytest.raises(service.Service.NoHost):
        s.ensure_non_empty_service_group()
    host = 'HOST'
    s = service.Service(tmpdir, settings.SHARE_DIR,
                        name=name,
                        host=host,
                        domain='test.com')
    hosts = s.ensure_non_empty_service_group()
    assert hosts == [host]
    assert hosts == s.ensure_non_empty_service_group()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2016, Peter Sagerson <psagers@ignorare.net>
# (c) 2016, Jiri Tyr <jiri.tyr@gmail.com>
# (c) 2017, Alexander Korinek <noles@a3k.net>
#
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils._text import to_native, to_bytes

import traceback
import re

try:
    import ldap
    import ldap.sasl

    HAS_LDAP = True
except ImportError:
    HAS_LDAP = False

__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = """
---
module: ldap_attrs
short_description: Add or remove multiple LDAP attribute values.
description:
  - Add or remove multiple LDAP attribute values.
notes:
  - This only deals with attributes on existing entries. To add or remove
    whole entries, see M(ldap_entry).
  - The default authentication settings will attempt to use a SASL EXTERNAL
    bind over a UNIX domain socket. This works well with the default Ubuntu
    install for example, which includes a cn=peercred,cn=external,cn=auth ACL
    rule allowing root to modify the server configuration. If you need to use
    a simple bind to access your server, pass the credentials in I(bind_dn)
    and I(bind_pw).
  - For I(state=present) and I(state=absent), all value comparisons are
    performed on the server for maximum accuracy. For I(state=exact), values
    have to be compared in Python, which obviously ignores LDAP matching
    rules. This should work out in most cases, but it is theoretically
    possible to see spurious changes when target and actual values are
    semantically identical but lexically distinct.
version_added: '2.5'
author:
  - Jiri Tyr (@jtyr)
  - Alexander Korinek (@noles)
requirements:
  - python-ldap
options:
  bind_dn:
    required: false
    default: null
    description:
      - A DN to bind with. If this is omitted, we'll try a SASL bind with
        the EXTERNAL mechanism. If this is blank, we'll use an anonymous
        bind.
  bind_pw:
    required: false
    default: null
    description:
      - The password to use with I(bind_dn).
  dn:
    required: true
    description:
      - The DN of the entry to modify.
  server_uri:
    required: false
    default: ldapi:///
    description:
      - A URI to the LDAP server. The default value lets the underlying
        LDAP client library look for a UNIX domain socket in its default
        location.
  start_tls:
    required: false
    choices: ['yes', 'no']
    default: 'no'
    description:
      - If true, we'll use the START_TLS LDAP extension.
  state:
    required: false
    choices: [present, absent, exact]
    default: present
    description:
      - The state of the attribute values. If C(present), all given
        values will be added if they're missing. If C(absent), all given
        values will be removed if present. If C(exact), the set of values
        will be forced to exactly those provided and no others. If
        I(state=exact) and I(value) is empty, all values for this
        attribute will be removed.
  attributes:
    required: true
    description:
      - The attribute(s) and value(s) to add or remove. The complex argument
        format is required in order to pass a list of strings (see examples).
  ordered:
    required: false
    choices: ['yes', 'no']
    default: 'no'
    description:
      - If C(yes), prepend list values with X-ORDERED index numbers in all
        attributes specified in the current task. This is useful mostly with
        I(olcAccess) attribute to easily manage LDAP Access Control Lists.
  validate_certs:
    required: false
    choices: ['yes', 'no']
    default: 'yes'
    description:
      - If C(no), SSL certificates will not be validated. This should only be
        used on sites using self-signed certificates.
"""


EXAMPLES = """
- name: Configure directory number 1 for example.com
  ldap_attrs:
    dn: olcDatabase={1}hdb,cn=config
    attributes:
        olcSuffix: dc=example,dc=com
    state: exact

# The complex argument format is required here to pass a list of ACL strings.
- name: Set up the ACL
  ldap_attrs:
    dn: olcDatabase={1}hdb,cn=config
    attributes:
        olcAccess:
          - >-
            {0}to attrs=userPassword,shadowLastChange
            by self write
            by anonymous auth
            by dn="cn=admin,dc=example,dc=com" write
            by * none'
          - >-
            {1}to dn.base="dc=example,dc=com"
            by dn="cn=admin,dc=example,dc=com" write
            by * read
    state: exact

# An alternative approach with automatic X-ORDERED numbering
- name: Set up the ACL
  ldap_attrs:
    dn: olcDatabase={1}hdb,cn=config
    attributes:
        olcAccess:
          - >-
            to attrs=userPassword,shadowLastChange
            by self write
            by anonymous auth
            by dn="cn=admin,dc=example,dc=com" write
            by * none'
          - >-
            to dn.base="dc=example,dc=com"
            by dn="cn=admin,dc=example,dc=com" write
            by * read
    ordered: yes
    state: exact

- name: Declare some indexes
  ldap_attrs:
    dn: olcDatabase={1}hdb,cn=config
    attributes:
        olcDbIndex:
            - objectClass eq
            - uid eq

- name: Set up a root user, which we can use later to bootstrap the directory
  ldap_attrs:
    dn: olcDatabase={1}hdb,cn=config
    attributes:
        olcRootDN: cn=root,dc=example,dc=com
        olcRootPW: "{SSHA}tabyipcHzhwESzRaGA7oQ/SDoBZQOGND"
    state: exact

- name: Get rid of an unneeded attribute
  ldap_attrs:
    dn: uid=jdoe,ou=people,dc=example,dc=com
    attributes:
        shadowExpire: ""
    state: exact
    server_uri: ldap://localhost/
    bind_dn: cn=admin,dc=example,dc=com
    bind_pw: password

#
# The same as in the previous example but with the authentication details
# stored in the ldap_auth variable:
#
# ldap_auth:
#   server_uri: ldap://localhost/
#   bind_dn: cn=admin,dc=example,dc=com
#   bind_pw: password
- name: Get rid of an unneeded attribute
  ldap_attrs:
    dn: uid=jdoe,ou=people,dc=example,dc=com
    attributes:
        shadowExpire: ""
    state: exact
    params: "{{ ldap_auth }}"
"""


RETURN = """
modlist:
  description: list of modified parameters
  returned: success
  type: list
  sample: '[[2, "olcRootDN", ["cn=root,dc=example,dc=com"]]]'
"""


class LdapAttr(object):
    def __init__(self, module):
        # Shortcuts
        self.module = module
        self.bind_dn = self.module.params['bind_dn']
        self.bind_pw = self.module.params['bind_pw']
        self.dn = self.module.params['dn']
        self.server_uri = self.module.params['server_uri']
        self.start_tls = self.module.params['start_tls']
        self.state = self.module.params['state']
        self.verify_cert = self.module.params['validate_certs']
        self.attrs = self.module.params['attributes']
        self.ordered = self.module.params['ordered']

        # Establish connection
        self.connection = self._connect_to_ldap()

    def _order_values(self, values):
        """ Preprend X-ORDERED index numbers to attribute's values. """
        ordered_values = []

        if isinstance(values, list):
            for index, value in enumerate(values):
                cleaned_value = re.sub(r'^\{\d+\}', '', value)
                ordered_values.append('{' + str(index) + '}' + cleaned_value)

        return ordered_values

    def _normalize_values(self, values):
        """ Normalize attribute's values. """
        norm_values = []

        if isinstance(values, list):
            if self.ordered:
                norm_values = list(map(to_bytes,
                                   self._order_values(list(map(str,
                                                               values)))))
            else:
                norm_values = list(map(to_bytes, values))
        elif values != "":
            norm_values = [to_bytes(str(values))]

        return norm_values

    def add(self):
        modlist = []
        for name, values in self.module.params['attributes'].items():
            norm_values = self._normalize_values(values)
            for value in norm_values:
                if self._is_value_absent(name, value):
                    modlist.append((ldap.MOD_ADD, name, value))

        return modlist

    def delete(self):
        modlist = []
        for name, values in self.module.params['attributes'].items():
            norm_values = self._normalize_values(values)
            for value in norm_values:
                if self._is_value_present(name, value):
                    modlist.append((ldap.MOD_DELETE, name, value))

        return modlist

    def exact(self):
        modlist = []
        for name, values in self.module.params['attributes'].items():
            norm_values = self._normalize_values(values)
            try:
                results = self.connection.search_s(
                    self.dn, ldap.SCOPE_BASE, attrlist=[name])
            except ldap.LDAPError as e:
                self.module.fail_json(
                    msg="Cannot search for attribute %s" % name,
                    details=to_native(e))

            current = results[0][1].get(name, [])

            if frozenset(norm_values) != frozenset(current):
                if len(current) == 0:
                    modlist.append((ldap.MOD_ADD, name, norm_values))
                elif len(norm_values) == 0:
                    modlist.append((ldap.MOD_DELETE, name, None))
                else:
                    modlist.append((ldap.MOD_REPLACE, name, norm_values))

        return modlist

    def _is_value_present(self, name, value):
        """ True if the target attribute has the given value. """
        try:
            is_present = bool(
                self.connection.compare_s(self.dn, name, value))
        except ldap.NO_SUCH_ATTRIBUTE:
            is_present = False

        return is_present

    def _is_value_absent(self, name, value):
        """ True if the target attribute doesn't have the given value. """
        return not self._is_value_present(name, value)

    def _connect_to_ldap(self):
        if not self.verify_cert:
            ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)

        connection = ldap.initialize(self.server_uri)

        if self.start_tls:
            try:
                connection.start_tls_s()
            except ldap.LDAPError as e:
                self.module.fail_json(msg="Cannot start TLS.",
                                      details=to_native(e))

        try:
            if self.bind_dn is not None:
                connection.simple_bind_s(self.bind_dn, self.bind_pw)
            else:
                connection.sasl_interactive_bind_s('', ldap.sasl.external())
        except ldap.LDAPError as e:
            self.module.fail_json(
                msg="Cannot bind to the server.", details=to_native(e))

        return connection


def main():
    module = AnsibleModule(
        argument_spec={
            'bind_dn': dict(default=None),
            'bind_pw': dict(default='', no_log=True),
            'dn': dict(required=True),
            'params': dict(type='dict'),
            'server_uri': dict(default='ldapi:///'),
            'start_tls': dict(default=False, type='bool'),
            'state': dict(
                default='present',
                choices=['present', 'absent', 'exact']),
            'attributes': dict(required=True, type='dict'),
            'ordered': dict(default=False, type='bool'),
            'validate_certs': dict(default=True, type='bool'),
        },
        supports_check_mode=True,
    )

    if not HAS_LDAP:
        module.fail_json(
            msg="Missing required 'ldap' module (pip install python-ldap)")

    # Update module parameters with user's parameters if defined
    if 'params' in module.params and isinstance(module.params['params'], dict):
        module.params.update(module.params['params'])
        # Remove the params
        module.params.pop('params', None)

    # Instantiate the LdapAttr object
    ldap = LdapAttr(module)

    state = module.params['state']

    # Perform action
    if state == 'present':
        modlist = ldap.add()
    elif state == 'absent':
        modlist = ldap.delete()
    elif state == 'exact':
        modlist = ldap.exact()

    changed = False

    if len(modlist) > 0:
        changed = True

        if not module.check_mode:
            try:
                ldap.connection.modify_s(ldap.dn, modlist)
            except Exception as e:
                module.fail_json(msg="Attribute action failed.",
                                 details=to_native(e),
                                 exception=traceback.format_exc())

    module.exit_json(changed=changed, modlist=modlist)


if __name__ == '__main__':
    main()



import yaml

from enough import settings


def get_driver():
    vars_dir = f'{settings.CONFIG_DIR}/inventory/group_vars/all'
    return yaml.safe_load(open(f'{vars_dir}/infrastructure.yml'))['infrastructure_driver']



from enough.common import retry

testinfra_hosts = ['ansible://chat-host']


def test_mattermost(host):
    with host.sudo():
        host.run("apt-get install -y curl")

    @retry.retry(AssertionError, tries=5)
    def run():
        r = host.run("curl -s -m 5 https://chat.$(hostname -d)")
        assert r.rc == 0
        assert 'Mattermost' in r.stdout

    run()



import os
import pytest
import sh
import shutil
import yaml

from enough import settings
from enough.common import ansible_utils


def test_get_global_variable():
    ansible = ansible_utils.Ansible(settings.CONFIG_DIR, settings.SHARE_DIR)
    assert ansible.get_global_variable('openstack_cloud')
    with pytest.raises(ansible_utils.AnsibleVariableNotFound):
        assert ansible.get_global_variable('UNKNOWN')


def test_get_hostvars():
    ansible = ansible_utils.Ansible(settings.CONFIG_DIR, settings.SHARE_DIR)
    d = ansible.get_hostvars('production_domain', 'bind-host',
                             'icinga-host')
    assert d == {'bind-host': 'enough.community', 'icinga-host': 'enough.community'}


def test_get_hostvars_recursive(tmpdir):
    """Check that we aren't affected by
    https://github.com/ansible/ansible/pull/64282"""

    host = 'test_get_hostvars_recursive_host'  # any host

    os.makedirs(f'{tmpdir}/group_vars/')
    groupvars = {
        'foo': 'blah',
        'bar': '{{ foo }}'
    }

    # An inventory file is required in order to avoid such warning:
    # [WARNING]: Unable to parse /tmp/pytest-0/test_get_hostvars_recursive0
    #            as an inventory source
    with open(f'{tmpdir}/hosts', 'w') as inventory:
        inventory.write(host)
    with open(f'{tmpdir}/group_vars/all.yaml', 'w') as group:
        yaml.dump(groupvars, group)
    ansible = ansible_utils.Ansible(settings.CONFIG_DIR, settings.SHARE_DIR,
                                    inventories=[tmpdir])

    r1 = ansible.get_hostvars(None, host)
    r2 = ansible.get_variable(None, host)

    assert r1[host]['bar'] == r2['bar'] == groupvars['foo']


def test_get_all_hostvars():
    ansible = ansible_utils.Ansible(settings.CONFIG_DIR, settings.SHARE_DIR)
    hostvars = ansible.get_hostvars(None)
    hosts = set(hostvars)

    # Check list of hosts isn't empty
    assert 'bind-host' in hosts
    assert 'icinga-host' in hosts

    inventory = ansible.ansible_inventory()
    inventory_vars = inventory['_meta']['hostvars']
    inventory_hosts = set(inventory_vars)

    # Check that output of ansible and ansible-inventory match
    assert hosts == inventory_hosts


def test_get_hostvars_unknown_host():
    ansible = ansible_utils.Ansible(settings.CONFIG_DIR, settings.SHARE_DIR)
    with pytest.raises(sh.ErrorReturnCode_1):
        ansible.get_hostvars('production_domain', 'inexistent-host')
    with pytest.raises(sh.ErrorReturnCode_1):
        ansible.get_hostvars(None, 'inexistent-host')


def test_get_hostvars_unknown_variable():
    ansible = ansible_utils.Ansible(settings.CONFIG_DIR, settings.SHARE_DIR)
    with pytest.raises(ansible_utils.AnsibleVariableNotFound):
        ansible.get_hostvars('inexistent-variable', 'bind-host')


def test_get_variable():
    test_inventory = 'tests/enough/common/test_ansible_utils/variables_inventory'
    expected = yaml.safe_load(open(f'{test_inventory}/host_vars/api-host.yml'))
    variable = 'api_admin_password'
    ansible = ansible_utils.Ansible(settings.CONFIG_DIR, settings.SHARE_DIR, [test_inventory])
    value = ansible.get_variable(variable, 'api-host')
    assert expected[variable] == value


def test_pipelining_is_enabled(monkeypatch):
    ansible = ansible_utils.Ansible(settings.CONFIG_DIR, settings.SHARE_DIR,
                                    ['localhost,'])
    ansible.bake()('all', '-llocalhost', '-massert',
                   "-athat='ansible_pipelining'")


def test_pipelining_is_enabled_2(monkeypatch):
    ansible = ansible_utils.Ansible(settings.CONFIG_DIR, settings.SHARE_DIR,
                                    ['localhost,'])

    # environment variable takes precedence over ansible.cfg
    monkeypatch.setenv('ANSIBLE_PIPELINING', 'False')
    with pytest.raises(sh.ErrorReturnCode):
        ansible.bake()('all', '-llocalhost', '-massert',
                       "-athat='ansible_pipelining'")


def test_get_role_variable():
    defaults = yaml.safe_load(open('playbooks/api/roles/api/defaults/main.yml'))
    variable = 'api_admin_password'
    playbook = ansible_utils.Playbook(settings.CONFIG_DIR, settings.SHARE_DIR)
    value = playbook.get_role_variable('api', variable, 'api-host')
    assert defaults[variable] == value


def test_get_role_unknown_variable():
    playbook = ansible_utils.Playbook(settings.CONFIG_DIR, settings.SHARE_DIR)
    with pytest.raises(sh.ErrorReturnCode_2):
        playbook.get_role_variable('api', 'inexistent_role_variable', 'api-host')


def test_get_role_variable_unknown_host():
    playbook = ansible_utils.Playbook(settings.CONFIG_DIR, settings.SHARE_DIR)
    with pytest.raises(sh.ErrorReturnCode_1):
        playbook.get_role_variable('api', 'api_admin_password', 'inexistent-host')


def test_get_unknown_role_variable():
    playbook = ansible_utils.Playbook(settings.CONFIG_DIR, settings.SHARE_DIR)
    with pytest.raises(sh.ErrorReturnCode_1):
        playbook.get_role_variable('inexistent-role', 'inexistent_role_variable', 'inexistent-host')


def test_playbook_roles_path():
    p = ansible_utils.Playbook(settings.CONFIG_DIR, settings.SHARE_DIR)
    r = p.roles_path('.')
    assert '/infrastructure/' in r


def test_playbook_ensure_decrypted(tmpdir):
    p = ansible_utils.Playbook(settings.CONFIG_DIR, settings.SHARE_DIR)
    shutil.copytree('tests/enough/common/test_ansible_utils/domain.com',
                    f'{tmpdir}/domain.com')
    c = f'{tmpdir}/domain.com'
    p.config_dir = c

    #
    # decryption is needed but no password is found
    #
    with pytest.raises(ansible_utils.Playbook.NoPasswordException):
        assert p.ensure_decrypted() is False
    shutil.copyfile('tests/enough/common/test_ansible_utils/domain.com.pass',
                    f'{tmpdir}/domain.com.pass')

    #
    # all files are decrypted
    #
    for f in p.encrypted_files(c):
        if f.endswith('not-encrypted.key'):
            continue
        assert p.is_encrypted(f)
    assert p.ensure_decrypted() is True
    for f in p.encrypted_files(c):
        assert not p.is_encrypted(f)
    #
    # nothing to do
    #
    assert p.ensure_decrypted() is False


def test_playbook_run_with_args(caplog):
    p = ansible_utils.Playbook(settings.CONFIG_DIR, settings.SHARE_DIR)
    p.run('tests/enough/common/test_ansible_utils/playbook-ok.yml')
    assert 'OK_PLAYBOOK' in caplog.text
    assert 'OK_IMPORTED' in caplog.text


def test_playbook_run_no_args(mocker):
    called = {}

    def playbook():
        def run(*args):
            assert '--private-key' in args
            called['playbook'] = True
        return run
    mocker.patch('enough.common.ansible_utils.Playbook.bake',
                 side_effect=playbook)
    kwargs = {
        'args': [],
    }
    p = ansible_utils.Playbook(settings.CONFIG_DIR, settings.SHARE_DIR)
    p.run_from_cli(**kwargs)
    assert 'playbook' in called


def test_ansible_inventory():
    i = ansible_utils.Ansible(settings.CONFIG_DIR, settings.SHARE_DIR).ansible_inventory()
    assert '_meta' in i


def test_get_groups():
    i = ansible_utils.Ansible(settings.CONFIG_DIR, settings.SHARE_DIR).get_groups()
    assert 'all-hosts' in i


def test_flat_inventory():
    inventory = {
        'group1': {
            'children': ['group2', 'group3'],
            'hosts': ['host10', 'host11'],
        },
        'group2': {
            'hosts': ['host20'],
            'children': ['group4'],
        },
        'group3': {
            'hosts': ['host30', 'host31'],
        },
        'group4': {
            'children': ['group5', 'group6'],
        },
        'group5': {
            'hosts': ['host50', 'host51'],
        },
        'group6': {},
        'group7': {
            'hosts': ['host70', 'host71'],
        }
    }
    ansible = ansible_utils.Ansible(settings.CONFIG_DIR, settings.SHARE_DIR)
    expected = ['host10', 'host11', 'host20', 'host30', 'host31', 'host50', 'host51']
    assert sorted(ansible._flat_inventory(inventory, inventory['group1'])) == expected
    expected = ['host70', 'host71']
    assert sorted(ansible._flat_inventory(inventory, inventory['group7'])) == expected



from tests.infrastructure import get_driver

testinfra_hosts = ['ansible://external-host']


def test_bind_external(host):
    bind_host = host.get_host('ansible://bind-host',
                              ansible_inventory=host.backend.ansible_inventory)
    domain = bind_host.run("hostname -d").stdout.strip()
    address = bind_host.ansible.get_variables()['ansible_host']

    with host.sudo():
        host.run("apt-get install -y dnsutils")

    infrastructure = get_driver()

    if infrastructure == 'openstack':
        cmd = host.run(f"dig ns1.{domain}")
        print(cmd.stdout)
        print(cmd.stderr)
        assert 0 == cmd.rc

    cmd = host.run(f"dig axfr {domain} @{address}")
    print(cmd.stdout)
    print(cmd.stderr)
    assert 0 == cmd.rc
    # recursion is prohibited
    cmd = host.run(f"dig fsf.org @{address} | grep -q '^fsf.org'")
    print(cmd.stdout)
    print(cmd.stderr)
    assert 1 == cmd.rc



from enough import cmd


def test_ssh(mocker, tmp_config_dir):
    # do not tamper with logging streams to avoid
    # ValueError: I/O operation on closed file.
    mocker.patch('cliff.app.App.configure_logging')
    mocker.patch('enough.common.ssh.SSH.ssh')
    assert cmd.main(['--debug', 'ssh', 'host']) == 0



# (c) 2015, Robert Chady <rchady@sitepen.com>
# Based on `runner/lookup_plugins/file.py` for Ansible
#   (c) 2012, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Debops.
# This file is NOT part of Ansible yet.
#
# Debops is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Debops.  If not, see <https://www.gnu.org/licenses/>.

import os

try:
    from debops import *
    from debops.cmds import *
except ImportError:
    pass

try:
    from ansible.plugins.lookup import LookupBase
except ImportError:
    LookupBase = object

from distutils.version import LooseVersion
from ansible import __version__ as __ansible_version__

'''

This file implements the `file_src` lookup filter for Ansible.  In difference
to the `file` filter, this searches values based on the `file-paths`
variable (colon separated) as configured in DebOps.

NOTE: This means this filter relies on DebOps.

'''

__author__ = "Robert Chady <rchady@sitepen.com>"
__copyright__ = "Copyright 2015 by Robert Chady <rchady@sitepen.com>"
__license__ = "GNU General Public LIcense version 3 (GPL v3) or later"

conf_tpl_paths = 'file-paths'


class LookupModule(LookupBase):

    def __new__(class_name, *args, **kwargs):
        if LooseVersion(__ansible_version__) < LooseVersion("2.0"):
            from ansible import utils, errors

            class LookupModuleV1(object):

                def __init__(self, basedir, *args, **kwargs):
                    self.basedir = basedir

                def run(self, terms, inject=None, **kwargs):

                    terms = utils.listify_lookup_plugin_terms(
                            terms, self.basedir, inject)
                    ret = []
                    config = {}
                    places = []

                    # this can happen if the variable contains a string,
                    # strictly not desired for lookup plugins, but users may
                    # try it, so make it work.
                    if not isinstance(terms, list):
                        terms = [terms]

                    try:
                        project_root = find_debops_project(required=False)
                        config = read_config(project_root)
                    except NameError:
                        pass

                    if ('paths' in config and
                            conf_tpl_paths in config['paths']):
                        custom_places = (
                                config['paths'][conf_tpl_paths].split(':'))
                        for custom_path in custom_places:
                            if os.path.isabs(custom_path):
                                places.append(custom_path)
                            else:
                                places.append(os.path.join(
                                    project_root, custom_path))

                    for term in terms:
                        if '_original_file' in inject:
                            relative_path = utils.path_dwim_relative(
                                    inject['_original_file'], 'files',
                                    '', self.basedir, check=False)
                            places.append(relative_path)
                        for path in places:
                            template = os.path.join(path, term)
                            if template and os.path.exists(template):
                                ret.append(template)
                                break
                        else:
                            raise errors.AnsibleError(
                                    "could not locate file in lookup: %s"
                                    % term)

                    return ret

            return LookupModuleV1(*args, **kwargs)

        else:
            from ansible.errors import AnsibleError
            from ansible.plugins.lookup import LookupBase

            class LookupModuleV2(LookupBase):

                def run(self, terms, variables=None, **kwargs):
                    ret = []
                    config = {}
                    places = []

                    # this can happen if the variable contains a string,
                    # strictly not desired for lookup plugins, but users may
                    # try it, so make it work.
                    if not isinstance(terms, list):
                        terms = [terms]

                    try:
                        project_root = find_debops_project(required=False)
                        config = read_config(project_root)
                    except NameError:
                        pass

                    if 'paths' in config and conf_tpl_paths in config['paths']:
                        custom_places = (
                                config['paths'][conf_tpl_paths].split(':'))
                        for custom_path in custom_places:
                            if os.path.isabs(custom_path):
                                places.append(custom_path)
                            else:
                                places.append(os.path.join(
                                    project_root, custom_path))

                    for term in terms:
                        if 'role_path' in variables:
                            relative_path = self._loader.path_dwim_relative(
                                    variables['role_path'], 'files', '')
                            places.append(relative_path)
                        for path in places:
                            template = os.path.join(path, term)
                            if template and os.path.exists(template):
                                ret.append(template)
                                break
                        else:
                            raise AnsibleError(
                                    "could not locate file in lookup: %s"
                                    % term)

                    return ret

            return LookupModuleV2(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
Support functions for command line utilities (scripts).
"""
# Copyright (C) 2014-2015 Hartmut Goebel <h.goebel@crazy-compilers.com>
# Part of the DebOps - https://debops.org/

# This program is free software; you can redistribute
# it and/or modify it under the terms of the
# GNU General Public License as published by the Free
# Software Foundation; either version 3 of the License,
# or (at your option) any later version.
#
# This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General
# Public License along with this program; if not,
# write to the Free Software Foundation, Inc., 59
# Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# An on-line copy of the GNU General Public License can
# be downloaded from the FSF web page at:
# https://www.gnu.org/copyleft/gpl.html

from __future__ import print_function

import os
import sys
import platform
import subprocess
try:
    from subprocess import DEVNULL  # py3k
except ImportError:
    # open DEVNULL like `subprocess` module does
    DEVNULL = os.open(os.devnull, os.O_RDWR)

from .. import find_debops_project as _find_debops_project, \
    find_monorepopath as _find_monorepopath, \
    find_playbookpath as _find_playbookpath, \
    find_inventorypath as _find_inventorypath

__author__ = "Hartmut Goebel <h.goebel@crazy-compilers.com>"
__copyright__ = "Copyright 2014-2015 by Hartmut Goebel "
"<h.goebel@crazy-compilers.com>"
__licence__ = "GNU General Public License version 3 (GPL v3) or later"


SCRIPT_NAME = os.path.basename(sys.argv[0])

# Don't check SSH fingerprint on connection (to enable, set INSECURE=1 on the
# command line)
INSECURE = bool(os.environ.get('INSECURE', False))

# External programms used. List here for easy substitution for
# hard-coded paths.
WHICH = 'which'


def error_msg(message, severity="Error"):
    """
    Display error message and exit
    """
    old = sys.stdout
    sys.stdout = sys.__stdout__
    print(SCRIPT_NAME+':', severity+':', message)
    sys.stdout = old
    if severity == "Error":
        raise SystemExit(1)


def require_commands(*cmd_names):
    """
    Check if required commands exist.
    """
    def command_exists(cmd_name):
        which = "where" if platform.system() == "Windows" else WHICH
        return not subprocess.call([which, cmd_name],
                                   stdout=DEVNULL, stderr=subprocess.STDOUT)

    for name in cmd_names:
        if not command_exists(name):
            error_msg("%s: command not found" % name)


def find_debops_project(path=None, required=True):
    project_root = _find_debops_project(path)
    if required and not project_root:
        # Exit if we are outside of project directory
        error_msg("Not a DebOps project directory")
    return project_root


def find_monorepopath(config, project_root, required=True):
    monorepo_path = _find_monorepopath(config, project_root)
    if required and not monorepo_path:
        error_msg("DebOps monorepo not installed")
    return monorepo_path


def find_playbookpath(config, project_root, required=True):
    playbooks_path = _find_playbookpath(config, project_root)
    if required and not playbooks_path:
        error_msg("DebOps playbooks not installed")
    return playbooks_path


def find_inventorypath(project_root, required=True):
    inventory = _find_inventorypath(project_root)
    if required and not inventory:
        error_msg("Ansible inventory not found")
    return inventory



import time
import requests
import yaml


def get_address(inventory):
    vars_dir = f'{inventory}/group_vars/all'
    return 'https://psono.' + yaml.safe_load(
        open(vars_dir + '/domain.yml'))['domain']


def test_psono(pytestconfig):
    # psono freshly recreated may take few mins to be operationnal
    url = get_address(pytestconfig.getoption("--ansible-inventory"))
    for i in range(60, 0, -1):
        r = requests.get(url, timeout=5, verify='certs')
        if r.status_code == requests.codes.ok:
            break
        time.sleep(5)
    assert 'psono' in r.text



import pytest

from enough.common.openstack import Heat, Stack
from enough.common.ssh import SSH

from tests import prepare_config_dir


@pytest.mark.openstack_integration
def test_ssh(openstack_name, openstack_provider, tmpdir):
    domain = 'enough.community'
    d = Heat.hostvars_to_stack(openstack_name, openstack_provider)
    config_dir = prepare_config_dir(domain, tmpdir)
    ssh = SSH(config_dir, domain=domain)
    s = Stack(config_dir, d)
    s.set_public_key(f'{config_dir}/infrastructure_key.pub')
    s.create_or_update()
    r = ssh.ssh(openstack_name, ['uptime'], interactive=False)
    s.delete()
    assert r.returncode == 0
    assert 'load average' in r.stdout.decode('utf-8')



from enough.common import retry

testinfra_hosts = ['ansible://cloud-host']


def test_nextcloud(host):
    with host.sudo():
        host.run("apt-get install -y curl")
    cmd = host.run("""
    set -xe
    curl --silent https://cloud.$(hostname -d)/login |
       grep --quiet 'This application requires JavaScript'
    """)
    print(cmd.stdout)
    print(cmd.stderr)
    assert 0 == cmd.rc


def test_nextcloud_via_tor(host):
    with host.sudo():
        host.run("apt-get install -y curl")

    @retry.retry(AssertionError, tries=5)
    def run():
        cmd = host.run("""
        set -xe
        hostname=$(sudo cat /var/lib/tor/services/cloud/hostname)
        torsocks curl --silent http://$hostname/login |
           grep --quiet 'This application requires JavaScript'
        """)
        print(cmd.stdout)
        print(cmd.stderr)
        assert 0 == cmd.rc

    run()
